"""Define config and fixtures for testing the functionality of isilon_hadoop_tools."""


from __future__ import absolute_import
from __future__ import unicode_literals

try:
    import configparser  # Python 3
except ImportError:
    import ConfigParser as configparser  # Python 2
from contextlib2 import ExitStack as does_not_raise
from enum import Enum
import json
import os
import random
import tempfile
try:
    from unittest.mock import Mock, patch  # Python 3
except ImportError:
    from mock import Mock, patch  # Python 2
import uuid

import kadmin
import pytest
import requests
import urllib3

from isilon_hadoop_tools import directories, identities, onefs, IsilonHadoopToolError


urllib3.disable_warnings()  # Without this, the SDK will emit InsecureRequestWarning on every call.


def pytest_addoption(parser):
    parser.addoption(
        '--address',
        help='OneFS Address',
    )
    parser.addoption(
        '--password',
        help='OneFS Admin Password',
    )
    parser.addoption(
        '--username',
        default='root',
        help='OneFS Admin Username',
    )
    parser.addoption(
        '--realm',
        help='Kerberos Realm',
    )
    parser.addoption(
        '--kadmin-address',
        help='Kerberos Administration Server Address',
    )
    parser.addoption(
        '--kdc-addresses',
        help='Kerberos Key Distribution Center Addresses',
        nargs='+',
    )
    parser.addoption(
        '--kadmin-username',
        help='Kerberos Administration Server Admin Username',
    )
    parser.addoption(
        '--kadmin-password',
        help='Kerberos Administration Server Admin Password',
    )


@pytest.fixture
def max_retry_exception_mock():
    """Get an object that raises MaxRetryError (from urllib3) when called."""
    return Mock(side_effect=urllib3.exceptions.MaxRetryError(pool=None, url=None))


@pytest.fixture(
    params=[
        'unresolvable.invalid',  # unresolvable
        'localhost',  # resolvable, not OneFS
        '127.0.0.1',  # IPv4, not OneFS
        '::1',  # IPv6, not OneFS -- If IPv6 is not enabled, this is the same as "unresolvable".
    ],
)
def invalid_address(request, max_retry_exception_mock):
    """Get an address that will cause connection errors for onefs.Client."""
    try:
        # This is how the SDK checks whether localhost is OneFS:
        # https://github.com/Isilon/isilon_sdk_python/blob/19958108ec550865ebeb1f2a4d250322cf4681c2/isi_sdk/rest.py#L33
        __import__('isi.rest')
    except ImportError:
        # Different hostnames/addresses hit errors in different code paths.
        # The first error that can be hit is a socket.gaierror if a hostname is unresolvable.
        # That won't get hit for addresses (e.g. 127.0.0.1 or ::1) or resolvable names, though.
        # Instead, those connections will succeed but will not respond correctly to API requests.
        # The first API request that's made is to get the cluster version (using isi_sdk_8_0).
        # To avoid having to wait for such a connection to time out, here we patch that request.
        with patch('isi_sdk_8_0.ClusterApi.get_cluster_version', max_retry_exception_mock):
            yield request.param  # yield to keep the patch until the teardown of the test.
    else:
        pytest.skip('Localhost is OneFS.')


@pytest.fixture(scope='session')
def onefs_client(pytestconfig):
    """Get an instance of onefs.Client."""
    return onefs.Client(
        address=pytestconfig.getoption('--address', skip=True),
        username=pytestconfig.getoption('--username', skip=True),
        password=pytestconfig.getoption('--password', skip=True),
        verify_ssl=False,  # OneFS uses a self-signed certificate by default.
    )


@pytest.fixture(scope='session')
def riptide_client(onefs_client):
    """Get an instance of onefs.Client that points to Riptide."""
    if onefs.ONEFS_RELEASES['8.0.0.0'] <= onefs_client.revision() < onefs.ONEFS_RELEASES['8.0.1.0']:
        return onefs_client
    pytest.skip('The OneFS cluster is not running Riptide.')


def new_name(request):
    """Get a name that may be used to create a new user or group."""
    return '-'.join([
        request.function.__name__,
        str(uuid.uuid4()),
    ])


def _new_group_name(request):
    return new_name(request)


@pytest.fixture
def new_group_name(request):
    """Get a name that may be used to create a new group."""
    return _new_group_name(request)


def new_id():
    """Get an ID that may be used to create a new user or group."""
    return random.randint(1024, 65536)


def _new_gid():
    return new_id()


@pytest.fixture
def new_gid():
    """Get a GID that may be used to create a new group."""
    return _new_gid()


def _deletable_group(request, onefs_client):
    name, gid = _new_group_name(request), _new_gid()
    onefs_client.create_group(name=name, gid=gid)
    return name, gid


@pytest.fixture
def deletable_group(request, onefs_client):
    """Get the name of an existing group that it is ok to delete."""
    return _deletable_group(request, onefs_client)


def _created_group(request, onefs_client):
    name, gid = _deletable_group(request, onefs_client)
    request.addfinalizer(lambda: onefs_client.delete_group(name='GID:' + str(gid)))
    return name, gid


@pytest.fixture
def created_group(request, onefs_client):
    """Get an existing group with a known GID."""
    return _created_group(request, onefs_client)


def _new_user_name(request):
    return new_name(request)


@pytest.fixture
def new_user_name(request):
    """Get a name that may be used to create a new user."""
    return _new_user_name(request)


def _new_uid():
    return new_id()


@pytest.fixture
def new_uid():
    """Get a UID that may be used to create a new user."""
    return _new_uid()


def _deletable_user(request, onefs_client):
    name = _new_user_name(request)
    primary_group_name, _ = _created_group(request, onefs_client)
    uid = _new_uid()
    onefs_client.create_user(name=name, primary_group_name=primary_group_name, uid=uid)
    return name, primary_group_name, uid


@pytest.fixture
def deletable_user(request, onefs_client):
    """Get the name of an existing user that it is ok to delete."""
    return _deletable_user(request, onefs_client)


def _created_user(request, onefs_client):
    name, primary_group_name, uid = _deletable_user(request, onefs_client)
    request.addfinalizer(lambda: onefs_client.delete_user(name='UID:' + str(uid)))
    return name, primary_group_name, uid


@pytest.fixture
def created_user(request, onefs_client):
    """Get an existing user with a known UID."""
    return _created_user(request, onefs_client)


def _deletable_proxy_user(request, onefs_client):
    user_name = _created_user(request, onefs_client)[0]
    members = []
    onefs_client.create_hdfs_proxy_user(name=user_name, members=members)
    return user_name, members


@pytest.fixture
def deletable_proxy_user(request, onefs_client):
    """Get the name of an existing proxy user that it is ok to delete."""
    return _deletable_proxy_user(request, onefs_client)


def _created_proxy_user(request, onefs_client):
    user_name, members = _deletable_proxy_user(request, onefs_client)
    request.addfinalizer(lambda: onefs_client.delete_hdfs_proxy_user(name=user_name))
    return user_name, members


@pytest.fixture
def created_proxy_user(request, onefs_client):
    """Get an existing proxy user with known members."""
    return _created_proxy_user(request, onefs_client)


def _deletable_realm(pytestconfig, onefs_client):
    realm = pytestconfig.getoption('--realm', skip=True)
    onefs_client.create_realm(
        name=realm,
        admin_server=pytestconfig.getoption('--kadmin-address', skip=True),
        kdcs=pytestconfig.getoption('--kdc-addresses', skip=True),
    )
    return realm


@pytest.fixture
def deletable_realm(pytestconfig, onefs_client):
    """Get the name of an existing realm that it is ok to delete."""
    return _deletable_realm(pytestconfig, onefs_client)


def _created_realm(request, onefs_client):
    realm = _deletable_realm(request.config, onefs_client)
    request.addfinalizer(lambda: onefs_client.delete_realm(name=realm))
    return realm


@pytest.fixture
def created_realm(request, onefs_client):
    """Get the name of an existing realm."""
    return _created_realm(request, onefs_client)


def _deletable_auth_provider(request, onefs_client):
    realm = _created_realm(request, onefs_client)
    onefs_client.create_auth_provider(
        realm=realm,
        user=request.config.getoption('--kadmin-username', skip=True),
        password=request.config.getoption('--kadmin-password', skip=True),
    )
    return realm


@pytest.fixture
def deletable_auth_provider(request, onefs_client):
    """Get the name of an existing Kerberos auth provider that it is ok to delete."""
    return _deletable_auth_provider(request, onefs_client)


def _created_auth_provider(request, onefs_client):
    auth_provider = _deletable_auth_provider(request, onefs_client)
    request.addfinalizer(lambda: onefs_client.delete_auth_provider(name=auth_provider))
    return auth_provider


@pytest.fixture
def created_auth_provider(request, onefs_client):
    """Get the name of an existing Kerberos auth provider."""
    return _created_auth_provider(request, onefs_client)


def _new_spn(request, onefs_client):
    return (
        _new_user_name(request) + '/' + onefs_client.address,
        _created_auth_provider(request, onefs_client),
    )


@pytest.fixture
def new_spn(request, onefs_client):
    """Get a principal that may be used to create a new SPN."""
    return _new_spn(request, onefs_client)


def _remove_principal_from_kdc(
        principal,
        realm,
        kdc,
        admin_server,
        admin_principal,
        admin_password,
):
    """Delete a Kerberos principal."""
    # Note: kadmin.init_with_password requires a Kerberos config file.

    # Create a temporary Kerberos config file.
    krb5_config = configparser.ConfigParser()
    krb5_config.optionxform = str
    krb5_config.add_section('libdefaults')
    krb5_config.set('libdefaults', 'default_realm', realm)
    krb5_config.add_section('realms')
    krb5_config.set(
        'realms',
        realm,
        '\n'.join([
            '{',
            '    kdc = ' + kdc,
            '    admin_server = ' + admin_server,
            '}',
        ]),
    )
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as krb5_conf:
        krb5_config.write(krb5_conf)

    # Activate the config file via an env var.
    previous_krb5_conf = os.environ.get('KRB5_CONFIG')
    os.environ['KRB5_CONFIG'] = krb5_conf.name

    # Delete the principal.
    kadmin.init_with_password(admin_principal, admin_password).delete_principal(principal)

    # Reset the env var.
    if previous_krb5_conf is None:
        del os.environ['KRB5_CONFIG']
    else:
        os.environ['KRB5_CONFIG'] = previous_krb5_conf

    # Delete the config file.
    os.remove(krb5_conf.name)


def _deletable_spn(request, onefs_client):
    spn, auth_provider = _new_spn(request, onefs_client)
    kadmin_username = request.config.getoption('--kadmin-username', skip=True)
    kadmin_password = request.config.getoption('--kadmin-password', skip=True)
    onefs_client.create_spn(
        spn=spn,
        realm=auth_provider,
        user=kadmin_username,
        password=kadmin_password,
    )
    request.addfinalizer(
        lambda: _remove_principal_from_kdc(
            principal=spn,
            realm=auth_provider,
            kdc=request.config.getoption('--kdc-addresses', skip=True)[0],
            admin_server=request.config.getoption('--kadmin-address', skip=True),
            admin_principal=kadmin_username,
            admin_password=kadmin_password,
        ),
    )
    return spn, auth_provider


@pytest.fixture
def deletable_spn(request, onefs_client):
    """Get the name of an existing SPN that it is ok to delete."""
    spn, auth_provider = _deletable_spn(request, onefs_client)
    yield spn, auth_provider
    assert (spn + '@' + auth_provider) not in onefs_client.list_spns(provider=auth_provider)


@pytest.fixture
def created_spn(request, onefs_client):
    """Get the name of an existing Kerberos SPN."""
    spn, auth_provider = _deletable_spn(request, onefs_client)
    request.addfinalizer(lambda: onefs_client.delete_spn(spn=spn, provider=auth_provider))
    return spn, auth_provider


@pytest.fixture
def exception():
    """Get an exception."""
    return random.choice([
        Exception,
        IsilonHadoopToolError,
    ])


def _api_exception_from_http_resp(onefs_client, body):
    return onefs_client._sdk.rest.ApiException(
        http_resp=urllib3.response.HTTPResponse(body=body),
    )


def _api_exception(onefs_client, messages=()):
    return _api_exception_from_http_resp(
        onefs_client,
        body=json.dumps({
            'errors': [
                {'message': message}
                for message in messages
            ],
        }),
    )


@pytest.fixture
def empty_api_exception_mock(onefs_client):
    """Get an object that raises an ApiException (from the Isilon SDK) when called."""
    return Mock(side_effect=_api_exception(onefs_client, messages=['']))


@pytest.fixture
def retriable_api_exception_mock(onefs_client):
    """Get an object that raises a retriable ApiException (from the Isilon SDK) when called."""
    return_value = None
    return (
        Mock(
            side_effect=[
                # First raise an exception, then return a value.
                _api_exception(onefs_client, messages=[onefs.APIError.try_again_error_format]),
                return_value,
            ],
        ),
        return_value,
    )


@pytest.fixture(
    params=[
        lambda onefs_client: (
            # body, decodable, valid, iterable, not empty, valid
            # This is known to occur in the wild.
            _api_exception(onefs_client, messages=[onefs.APIError.try_again_error_format]),
            does_not_raise(),
        ),
        lambda onefs_client: (
            # body, decodable, valid, iterable, not empty, invalid (KeyError)
            _api_exception_from_http_resp(onefs_client, body='{"errors": [{}]}'),
            pytest.raises(onefs.MalformedAPIError),
        ),
        lambda onefs_client: (
            # body, decodable, valid, iterable, not empty, invalid (TypeError)
            _api_exception_from_http_resp(onefs_client, body='{"errors": [[]]}'),
            pytest.raises(onefs.MalformedAPIError),
        ),
        lambda onefs_client: (
            # body, decodable, valid, iterable, empty
            _api_exception_from_http_resp(onefs_client, body='{"errors": []}'),
            does_not_raise(),
        ),
        lambda onefs_client: (
            # body, decodable, valid, not iterable
            _api_exception_from_http_resp(onefs_client, body='{"errors": null}'),
            pytest.raises(onefs.MalformedAPIError),
        ),
        lambda onefs_client: (
            # body, decodable, invalid (KeyError)
            # This is known to occur in the wild (e.g. bug 248011)
            _api_exception_from_http_resp(onefs_client, body='{}'),
            pytest.raises(onefs.MalformedAPIError),
        ),
        lambda onefs_client: (
            # body, decodable, invalid (TypeError)
            _api_exception_from_http_resp(onefs_client, body='[]'),
            pytest.raises(onefs.MalformedAPIError),
        ),
        lambda onefs_client: (
            # body, undecodable
            # This is known to occur in the wild (e.g. if Apache errors before PAPI).
            _api_exception_from_http_resp(onefs_client, body='not JSON'),
            pytest.raises(onefs.UndecodableAPIError),
        ),
        lambda onefs_client: (
            # no body
            # This is known to occur in the wild.
            onefs_client._sdk.rest.ApiException(status=0, reason='built without http_resp'),
            pytest.raises(onefs.UndecodableAPIError),
        ),
        lambda onefs_client: (
            # uninitialized
            onefs_client._sdk.rest.ApiException(),
            pytest.raises(onefs.UndecodableAPIError),
        ),
    ],
)
def api_error_errors_expectation(request, onefs_client):
    """
    Get an APIError and the expectation (context manager)
    of what happens when the errors method is called.
    """
    api_exception, expectation = request.param(onefs_client)
    return onefs.APIError(api_exception), expectation


@pytest.fixture
def api_error(api_error_errors_expectation):
    """Get a onefs.APIError exception."""
    return api_error_errors_expectation[0]


MAX_MODE = 0o1777


@pytest.fixture
def max_mode():
    """Get the highest integer mode this test suite will use."""
    return MAX_MODE


def _deletable_directory(request, onefs_client):
    path = '/' + new_name(request)
    mode = random.randint(0, MAX_MODE)
    mode &= 0o777  # https://bugs.west.isilon.com/show_bug.cgi?id=250615
    onefs_client.mkdir(path=path, mode=mode)
    return path, {
        'group': onefs_client.primary_group_of_user(onefs_client.username),
        'mode': mode,
        'owner': onefs_client.username,
    }


@pytest.fixture
def deletable_directory(request, onefs_client):
    """Get the path and mode of an existing directory that it is ok to delete."""
    return _deletable_directory(request, onefs_client)


def _created_directory(request, onefs_client):
    path, permissions = _deletable_directory(request, onefs_client)
    request.addfinalizer(lambda: onefs_client.rmdir(path=path, recursive=True))
    return path, permissions


@pytest.fixture
def created_directory(request, onefs_client):
    """Get an existing directory with a known mode."""
    return _created_directory(request, onefs_client)


@pytest.fixture
def supported_feature():
    """Get a OneFSFeature that is guaranteed to be supported."""
    return onefs.OneFSFeature.FOREVER


@pytest.fixture
def unsupported_feature():
    """Get a OneFSFeature that is guaranteed to be unsupported."""
    class OneFSFakeFeature(Enum):
        FAKE_FEATURE = (float('inf'), 0)
    return OneFSFakeFeature.FAKE_FEATURE


@pytest.fixture
def requests_delete_raises():
    class _DummyResponse(object):
        def raise_for_status(self):
            raise requests.exceptions.HTTPError
    with patch('requests.delete', lambda *args, **kwargs: _DummyResponse()):
        yield


@pytest.fixture(params=['cdh', 'cdp', 'hdp'])
def users_groups_for_directories(request, onefs_client):
    """
    Get users and groups from the identities module that
    correspond to directories from the directories module
    (i.e. get the identities guaranteed to exist for a set of directories).
    """

    users, groups = set(), set()

    def _pass(*args, **kwargs):
        pass

    identities.iterate_identities(
        {
            'cdh': identities.cdh_identities,
            'cdp': identities.cdh_identities,
            'hdp': identities.hdp_identities,
        }[request.param](onefs_client.zone),
        create_group=lambda group_name: groups.add(group_name),
        create_user=lambda user_name, _: users.add(user_name),
        add_user_to_group=_pass,
        create_proxy_user=_pass,
    )
    return (
        (users, groups),
        {
            'cdh': directories.cdh_directories,
            'cdp': directories.cdh_directories,
            'hdp': directories.hdp_directories,
        }[request.param]()
    )



import posixpath
import subprocess
import uuid

import pytest


@pytest.fixture
def empty_hdfs_root(onefs_client):
    """Create a temporary directory and make it the HDFS root."""
    old_hdfs_root = onefs_client.hdfs_settings()['root_directory']
    new_root_name = str(uuid.uuid4())
    onefs_client.mkdir(new_root_name, 0o755)
    onefs_client.update_hdfs_settings({
        'root_directory': posixpath.join(onefs_client.zone_settings()['path'], new_root_name),
    })
    yield
    onefs_client.update_hdfs_settings({'root_directory': old_hdfs_root})
    onefs_client.rmdir(new_root_name, recursive=True)


@pytest.mark.usefixtures('empty_hdfs_root')
@pytest.mark.parametrize('script', ['isilon_create_users', 'isilon_create_directories'])
@pytest.mark.parametrize('dist', ['cdh', 'cdp', 'hdp'])
def test_dry_run(script, onefs_client, dist):
    subprocess.check_call([
        script,
        '--append-cluster-name', str(uuid.uuid4()),
        '--dist', dist,
        '--dry',
        '--no-verify',
        '--onefs-password', onefs_client.password,
        '--onefs-user', onefs_client.username,
        '--zone', 'System',
        onefs_client.address,
    ])



"""Verify the functionality of isilon_hadoop_tools.__init__."""


from __future__ import absolute_import
from __future__ import unicode_literals

import pytest

import isilon_hadoop_tools


@pytest.mark.parametrize(
    'error, classinfo',
    [
        (isilon_hadoop_tools.IsilonHadoopToolError, Exception),
    ],
)
def test_errors(error, classinfo):
    """Ensure that exception types remain consistent."""
    assert issubclass(error, isilon_hadoop_tools.IsilonHadoopToolError)
    assert issubclass(error, classinfo)



import pytest

from isilon_hadoop_tools import IsilonHadoopToolError, directories


def test_directory_identities(users_groups_for_directories):
    """
    Verify that identities needed by the directories module
    are guaranteed to exist by the identities module.
    """
    (users, groups), dirs = users_groups_for_directories
    for hdfs_directory in dirs:
        assert hdfs_directory.owner in users
        assert hdfs_directory.group in groups


@pytest.mark.parametrize(
    'error, classinfo',
    [
        (directories.DirectoriesError, IsilonHadoopToolError),
        (directories.HDFSRootDirectoryError, directories.DirectoriesError),
    ],
)
def test_errors_cli(error, classinfo):
    """Ensure that exception types remain consistent."""
    assert issubclass(error, IsilonHadoopToolError)
    assert issubclass(error, directories.DirectoriesError)
    assert issubclass(error, classinfo)



"""This module defines a CLI common to all command-line tools."""

from __future__ import absolute_import
from __future__ import unicode_literals

import argparse
import getpass
import logging

from future.utils import raise_from

import isilon_hadoop_tools.onefs


__all__ = [
    # Decorators
    'catches',

    # Exceptions
    'CLIError',
    'HintedError',

    # Functions
    'base_cli',
    'configure_logging',
    'hdfs_client',
    'logging_cli',
    'onefs_cli',
    'onefs_client',
]

LOGGER = logging.getLogger(__name__)


class CLIError(isilon_hadoop_tools.IsilonHadoopToolError):
    """All Exceptions emitted from this module inherit from this Exception."""


def catches(exception):
    """Create a decorator for functions that emit the specified exception."""
    def decorator(func):
        """Decorate a function that should catch instances of the specified exception."""
        def decorated(*args, **kwargs):
            """Catch instances of a specified exception that are raised from the function."""
            try:
                return func(*args, **kwargs)
            except exception as ex:
                logging.error(ex)
                return 1
        return decorated
    return decorator


def base_cli(parser=None):
    """Define common CLI arguments and options."""
    if parser is None:
        parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    onefs_cli(parser.add_argument_group('OneFS'))
    logging_cli(parser.add_argument_group('Logging'))
    return parser


def onefs_cli(parser=None):
    """Define OneFS CLI arguments and options."""
    if parser is None:
        parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "--zone", "-z",
        help="Specify a OneFS access zone.",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--no-verify",
        help="Do not verify SSL/TLS certificates.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--onefs-password",
        help="Specify the password for --onefs-user.",
        type=str,
    )
    parser.add_argument(
        "--onefs-user",
        help="Specify the user to connect to OneFS as.",
        type=str,
        default="root",
    )
    parser.add_argument(
        "onefs_address",
        help="Specify an IP address or FQDN/SmartConnect that "
        "can be used to connect to and configure OneFS.",
        type=str,
    )
    return parser


class HintedError(CLIError):

    """
    This exception is used to modify the error message passed to the user
    when a common error occurs that has a possible solution the user will likely want.
    """

    def __str__(self):
        base_str = super(HintedError, self).__str__()
        return str(getattr(self, '__cause__', None)) + '\nHint: ' + base_str


def _client_from_onefs_cli(init, args):
    try:
        return init(
            address=args.onefs_address,
            username=args.onefs_user,
            password=getpass.getpass() if args.onefs_password is None else args.onefs_password,
            default_zone=args.zone,
            verify_ssl=not args.no_verify,
        )
    except isilon_hadoop_tools.onefs.OneFSCertificateError as exc:
        raise_from(
            HintedError('--no-verify can be used to skip certificate verification.'),
            exc,
        )
    except isilon_hadoop_tools.onefs.MissingLicenseError as exc:
        raise_from(
            CLIError(
                (
                    isilon_hadoop_tools.onefs.APIError.license_expired_error_format
                    if isinstance(exc, isilon_hadoop_tools.onefs.ExpiredLicenseError) else
                    isilon_hadoop_tools.onefs.APIError.license_missing_error_format
                ).format(exc),
            ),
            exc,
        )
    except isilon_hadoop_tools.onefs.MissingZoneError as exc:
        raise_from(
            CLIError(isilon_hadoop_tools.onefs.APIError.zone_not_found_error_format.format(exc)),
            exc,
        )


def hdfs_client(args):
    """Get a onefs.Client.for_hdfs from args parsed by onefs_cli."""
    return _client_from_onefs_cli(isilon_hadoop_tools.onefs.Client.for_hdfs, args)


def onefs_client(args):
    """Get a onefs.Client from args parsed by onefs_cli."""
    return _client_from_onefs_cli(isilon_hadoop_tools.onefs.Client, args)


def logging_cli(parser=None):
    """Define logging CLI arguments and options."""
    if parser is None:
        parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-q', '--quiet',
        default=False,
        action='store_true',
        help='Supress console output.',
    )
    parser.add_argument(
        "--log-file",
        type=str,
        help="Specify a path to log to.",
    )
    parser.add_argument(
        "--log-level",
        help="Specify how verbose logging should be.",
        default='info',
        choices=('debug', 'info', 'warning', 'error', 'critical'),
    )
    return parser


def configure_logging(args):
    """Configure logging for command-line tools."""
    logging.getLogger().setLevel(logging.getLevelName(args.log_level.upper()))
    if not args.quiet:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(logging.Formatter('[%(levelname)s] %(message)s'))
        logging.getLogger().addHandler(console_handler)
    if args.log_file:
        logfile_handler = logging.FileHandler(args.log_file)
        logfile_handler.setFormatter(
            logging.Formatter('[%(asctime)s] %(name)s [%(levelname)s] %(message)s'),
        )
        logging.getLogger().addHandler(logfile_handler)



"""Command-line interface for entry points"""

from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import os
import sys
import time

from future.utils import raise_from
import urllib3

import isilon_hadoop_tools
import isilon_hadoop_tools.cli
import isilon_hadoop_tools.directories
import isilon_hadoop_tools.identities


DRY_RUN = 'Had this been for real, this is what would have happened...'
LOGGER = logging.getLogger(__name__)


def base_cli(parser=None):
    """Define CLI arguments and options for all entry points."""
    if parser is None:
        parser = isilon_hadoop_tools.cli.base_cli()
    parser.add_argument(
        '--append-cluster-name',
        help='the cluster name to append on identities',
        type=str,
    )
    parser.add_argument(
        '--dist',
        help='the Hadoop distribution to be deployed',
        choices=('cdh', 'cdp', 'hdp'),
        required=True,
    )
    parser.add_argument(
        '--dry',
        help='do a dry run (only logs)',
        action='store_true',
        default=False,
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s v{0}'.format(isilon_hadoop_tools.__version__),
    )
    return parser


def configure_script(args):
    """Logic that applies to all scripts goes here."""
    if args.no_verify:
        urllib3.disable_warnings()


def isilon_create_users_cli(parser=None):
    """Define CLI arguments and options for isilon_create_users."""
    if parser is None:
        parser = base_cli()
    parser.add_argument(
        '--start-gid',
        help='the lowest GID to create a group with',
        type=int,
        default=isilon_hadoop_tools.identities.Creator.default_start_gid,
    )
    parser.add_argument(
        '--start-uid',
        help='the lowest UID to create a user with',
        type=int,
        default=isilon_hadoop_tools.identities.Creator.default_start_uid,
    )
    return parser


@isilon_hadoop_tools.cli.catches(isilon_hadoop_tools.IsilonHadoopToolError)
def isilon_create_users(argv=None):
    """Execute isilon_create_users commands."""

    if argv is None:
        argv = sys.argv[1:]
    args = isilon_create_users_cli().parse_args(argv)

    isilon_hadoop_tools.cli.configure_logging(args)
    configure_script(args)
    onefs = isilon_hadoop_tools.cli.hdfs_client(args)

    identities = {
        'cdh': isilon_hadoop_tools.identities.cdh_identities,
        'cdp': isilon_hadoop_tools.identities.cdp_identities,
        'hdp': isilon_hadoop_tools.identities.hdp_identities,
    }[args.dist](args.zone)

    name = '-'.join([
        str(int(time.time())),
        args.zone,
        args.dist,
    ])

    if args.append_cluster_name is not None:
        suffix = args.append_cluster_name
        if not suffix.startswith('-'):
            suffix = '-' + suffix
        identities = isilon_hadoop_tools.identities.with_suffix_applied(identities, suffix)
        name += suffix

    onefs_and_files = isilon_hadoop_tools.identities.Creator(
        onefs=onefs,
        onefs_zone=args.zone,
        start_uid=args.start_uid,
        start_gid=args.start_gid,
        script_path=os.path.join(os.getcwd(), name + '.sh'),
    )
    if args.dry:
        LOGGER.info(DRY_RUN)
        LOGGER.info('A script would have been created at %s.', onefs_and_files.script_path)
        LOGGER.info('The following actions would have populated it and OneFS:')
        onefs_and_files.log_identities(identities)
    else:
        onefs_and_files.create_identities(identities)


def isilon_create_directories_cli(parser=None):
    """Define CLI arguments and options for isilon_create_directories."""
    if parser is None:
        parser = base_cli()
    return parser


@isilon_hadoop_tools.cli.catches(isilon_hadoop_tools.IsilonHadoopToolError)
def isilon_create_directories(argv=None):
    """Execute isilon_create_directories commands."""

    if argv is None:
        argv = sys.argv[1:]
    args = isilon_create_directories_cli().parse_args(argv)

    isilon_hadoop_tools.cli.configure_logging(args)
    configure_script(args)
    onefs = isilon_hadoop_tools.cli.hdfs_client(args)

    suffix = args.append_cluster_name
    if suffix is not None and not suffix.startswith('-'):
        suffix = '-' + suffix

    directories = {
        'cdh': isilon_hadoop_tools.directories.cdh_directories,
        'cdp': isilon_hadoop_tools.directories.cdp_directories,
        'hdp': isilon_hadoop_tools.directories.hdp_directories,
    }[args.dist](identity_suffix=suffix)

    creator = isilon_hadoop_tools.directories.Creator(
        onefs=onefs,
        onefs_zone=args.zone,
    )
    try:
        if args.dry:
            LOGGER.info(DRY_RUN)
            creator.log_directories(directories)
        else:
            creator.create_directories(directories)
    except isilon_hadoop_tools.directories.HDFSRootDirectoryError as exc:
        raise_from(
            isilon_hadoop_tools.cli.CLIError(
                'The HDFS root directory must not be {0}.'.format(exc),
            ),
            exc,
        )



"""Define and create necessary Hadoop users and groups on OneFS."""

from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import os

import isilon_hadoop_tools.onefs


__all__ = [
    # Functions
    'cdh_identities',
    'cdp_identities',
    'hdp_identities',
    'iterate_identities',
    'log_identities',
    'with_suffix_applied',

    # Objects
    'Creator',
]

LOGGER = logging.getLogger(__name__)


def _log_create_group(group_name):
    LOGGER.info('Create %s group.', group_name)


def _log_create_user(user_name, pgroup_name):
    LOGGER.info('Create %s:%s user.', user_name, pgroup_name)


def _log_add_user_to_group(user_name, group_name):
    LOGGER.info('Add %s user to %s group.', user_name, group_name)


def _log_create_proxy_user(proxy_user_name, members):
    LOGGER.info(
        'Create %s proxy user with the following members: %s.',
        proxy_user_name,
        ', '.join(
            '{0} ({1})'.format(member_name, member_type)
            for member_name, member_type in members
        ),
    )


class Creator(object):

    """
    Create users and groups with contiguous IDs on OneFS
    and in a local user/group creation script for Linux.
    """
    # pylint: disable=logging-format-interpolation

    default_start_uid = 1025
    default_start_gid = 1025

    def __init__(  # pylint: disable=too-many-arguments
            self,
            onefs,
            onefs_zone=None,
            start_uid=default_start_uid,
            start_gid=default_start_gid,
            script_path=None,
    ):
        self.onefs = onefs
        self.onefs_zone = onefs_zone
        self._next_uid = start_uid
        self._next_gid = start_gid
        self.script_path = script_path

    @property
    def next_gid(self):
        """Get the next monotonically-increasing GID (begins at start_gid)."""
        try:
            return self._next_gid
        finally:
            self._next_gid += 1

    @property
    def next_uid(self):
        """Get the next monotonically-increasing UID (begins at start_uid)."""
        try:
            return self._next_uid
        finally:
            self._next_uid += 1

    def add_user_to_group(self, user_name, group_name):
        """Add a user to a group on OneFS and in the local group-creation script."""
        try:
            LOGGER.info(
                'Adding the %s user to the %s group on %s...',
                user_name,
                group_name,
                self.onefs.address,
            )
            self.onefs.add_user_to_group(
                user_name=user_name,
                group_name=group_name,
                zone=self.onefs_zone,
            )
        except isilon_hadoop_tools.onefs.APIError as exc:
            uid = self.onefs.uid_of_user(user_name=user_name, zone=self.onefs_zone)
            if exc.user_already_in_group_error(uid, group_name):
                LOGGER.warning(exc.user_already_in_group_error_format.format(uid, group_name))
            else:
                raise
        if self.script_path:
            self._create_script()
            LOGGER.info(
                'Adding the %s user to the %s group in %s...',
                user_name,
                group_name,
                self.script_path,
            )
            with open(self.script_path, 'a') as script_file:
                script_file.write(
                    'usermod -a -G {group} {user}\n'.format(group=group_name, user=user_name),
                )

    def create_group(self, group_name):
        """Create a group on OneFS and in the local script."""
        while True:
            try:
                gid = self.next_gid
                LOGGER.info(
                    'Creating the %s group with GID %s on %s...',
                    group_name,
                    gid,
                    self.onefs.address,
                )
                self.onefs.create_group(name=group_name, gid=gid, zone=self.onefs_zone)
                break
            except isilon_hadoop_tools.onefs.APIError as exc:
                if exc.gid_already_exists_error(gid):
                    LOGGER.warning(exc.gid_already_exists_error_format.format(gid))
                    continue
                if exc.group_already_exists_error(group_name):
                    LOGGER.warning(exc.group_already_exists_error_format.format(group_name))
                    gid = self.onefs.gid_of_group(group_name=group_name, zone=self.onefs_zone)
                    break
                raise
        if self.script_path:
            self._create_script()
            LOGGER.info(
                'Creating the %s group with GID %s in %s...',
                group_name,
                gid,
                self.script_path,
            )
            with open(self.script_path, 'a') as script_file:
                script_file.write('groupadd --gid {gid} {name}\n'.format(gid=gid, name=group_name))
        return gid

    def create_identities(
            self,
            identities,
            create_group=None,
            create_user=None,
            add_user_to_group=None,
            create_proxy_user=None,
            _flush_auth_cache=None,
            _create_script=None,
    ):
        """Create identities on OneFS and in the local script."""
        if self.onefs_zone.lower() == 'system':
            LOGGER.warning('Deploying in the System zone is not recommended.')
        if self.script_path:
            LOGGER.info('Creating %s...', self.script_path)
            (_create_script or self._create_script)()
        iterate_identities(
            identities,
            create_group=create_group or self.create_group,
            create_user=create_user or self.create_user,
            add_user_to_group=add_user_to_group or self.add_user_to_group,
            create_proxy_user=create_proxy_user or self.create_proxy_user,
        )
        LOGGER.info('Flushing the auth cache...')
        (_flush_auth_cache or self.onefs.flush_auth_cache)()

    def log_identities(self, identities):
        """Log the actions that would be taken by create_identities."""
        self.create_identities(
            identities,
            create_group=_log_create_group,
            create_user=_log_create_user,
            add_user_to_group=_log_add_user_to_group,
            create_proxy_user=_log_create_proxy_user,
            _flush_auth_cache=lambda: None,
            _create_script=lambda: None,
        )

    def create_proxy_user(self, proxy_user_name, members):
        """Create a proxy user on OneFS."""
        try:
            LOGGER.info(
                'Creating the %s proxy user with the following members: %s...',
                proxy_user_name,
                ', '.join(
                    '{0} ({1})'.format(member_name, member_type)
                    for member_name, member_type in members
                ),
            )
            self.onefs.create_hdfs_proxy_user(
                name=proxy_user_name,
                members=members,
                zone=self.onefs_zone,
            )
        except isilon_hadoop_tools.onefs.APIError as exc:
            if exc.proxy_user_already_exists_error(proxy_user_name=proxy_user_name):
                LOGGER.warning(exc.proxy_user_already_exists_error_format.format(proxy_user_name))
                return
            raise

    def _create_script(self):
        if not os.path.exists(self.script_path):
            with open(self.script_path, 'w') as script_file:
                script_file.write('#!/usr/bin/env sh\n')
                script_file.write('set -o errexit\n')
                script_file.write('set -o xtrace\n')

    def create_user(self, user_name, primary_group_name):
        """Create a user on OneFS and in the local script."""
        while True:
            try:
                uid = self.next_uid
                LOGGER.info(
                    'Creating the %s user with UID %s on %s...',
                    user_name,
                    uid,
                    self.onefs.address,
                )
                self.onefs.create_user(
                    name=user_name,
                    uid=uid,
                    primary_group_name=primary_group_name,
                    zone=self.onefs_zone,
                    enabled=True,
                )
                break
            except isilon_hadoop_tools.onefs.APIError as exc:
                if exc.uid_already_exists_error(uid):
                    LOGGER.warning(exc.uid_already_exists_error_format.format(uid))
                    continue
                if exc.user_already_exists_error(user_name):
                    LOGGER.warning(exc.user_already_exists_error_format.format(user_name))
                    uid = self.onefs.uid_of_user(user_name=user_name, zone=self.onefs_zone)
                    break
                raise
        if self.script_path:
            self._create_script()
            LOGGER.info(
                'Creating the %s user with UID %s in %s...',
                user_name,
                uid,
                self.script_path,
            )
            with open(self.script_path, 'a') as script_file:
                script_file.write(
                    'useradd --uid {uid} --gid {gid} {name}\n'.format(
                        uid=uid,
                        gid=self.onefs.gid_of_group(
                            group_name=self.onefs.primary_group_of_user(
                                user_name=user_name,
                                zone=self.onefs_zone,
                            ),
                            zone=self.onefs_zone,
                        ),
                        name=user_name,
                    ),
                )
        return uid


def iterate_identities(
        identities,
        create_group,
        create_user,
        add_user_to_group,
        create_proxy_user,
):
    """Iterate over all groups, users, and proxy users in creation-order."""

    created_group_names = set()
    for group_name in identities['groups']:
        if group_name not in created_group_names:
            create_group(group_name)
            created_group_names.add(group_name)

    for user_name, (pgroup_name, sgroup_names) in identities['users'].items():
        for group_name in sgroup_names.union({pgroup_name}):
            if group_name not in created_group_names:
                create_group(group_name)
                created_group_names.add(group_name)
        create_user(user_name, pgroup_name)
        for group_name in sgroup_names:
            add_user_to_group(user_name, group_name)

    for proxy_user_name, members in identities['proxy_users'].items():
        create_proxy_user(proxy_user_name, members)


def log_identities(identities):
    """Iterate identities in creation-order and log the actions that would be taken."""
    iterate_identities(
        identities,
        create_group=_log_create_group,
        create_user=_log_create_user,
        add_user_to_group=_log_add_user_to_group,
        create_proxy_user=_log_create_proxy_user,
    )


def with_suffix_applied(
        identities,
        suffix,
        applicator=lambda identity, suffix: identity + suffix,
):
    """Append a suffix to all identities."""
    return {
        'groups': {applicator(group_name, suffix) for group_name in identities['groups']},
        'users': {
            applicator(user_name, suffix): (
                applicator(pgroup_name, suffix),
                {applicator(sgroup_name, suffix) for sgroup_name in sgroup_names},
            )
            for user_name, (pgroup_name, sgroup_names) in identities['users'].items()
        },
        'proxy_users': {
            applicator(proxy_user_name, suffix): {
                (applicator(member_name, suffix), member_type)
                for member_name, member_type in members
            }
            for proxy_user_name, members in identities['proxy_users'].items()
        }
    }


def cdh_identities(zone):
    """Identities needed for Cloudera Distribution including Hadoop"""
    smoke_user = ('cloudera-scm', 'user')
    identities = {
        'groups': set(),  # Groups with no users in them.
        'users': {
            'accumulo': ('accumulo', set()),
            'anonymous': ('anonymous', set()),
            'apache': ('apache', set()),
            'cloudera-scm': ('cloudera-scm', set()),
            'cmjobuser': ('cmjobuser', set()),
            'flume': ('flume', set()),
            'hbase': ('hbase', {'hadoop', 'supergroup'}),
            'hdfs': ('hdfs', {'hadoop', 'supergroup'}),
            'hive': ('hive', set()),
            'HTTP': ('HTTP', {'hadoop', 'supergroup'}),
            'httpfs': ('httpfs', set()),
            'hue': ('hue', set()),
            'impala': ('impala', {'hive'}),
            'kafka': ('kafka', set()),
            'keytrustee': ('keytrustee', set()),
            'kms': ('kms', set()),
            'kudu': ('kudu', set()),
            'llama': ('llama', set()),
            'mapred': ('mapred', {'hadoop', 'supergroup'}),
            'oozie': ('oozie', set()),
            'sentry': ('sentry', set()),
            'solr': ('solr', set()),
            'spark': ('spark', set()),
            'sqoop': ('sqoop', {'sqoop2'}),
            'sqoop2': ('sqoop2', {'sqoop'}),
            'yarn': ('yarn', {'hadoop', 'supergroup'}),
            'zookeeper': ('zookeeper', set()),
        },
        'proxy_users': {
            'flume': {smoke_user, ('hadoop', 'group')},
            'hive': {smoke_user, ('hadoop', 'group')},
            'HTTP': {smoke_user},
            'hue': {smoke_user, ('hadoop', 'group')},
            'impala': {smoke_user, ('hadoop', 'group')},
            'mapred': {smoke_user, ('hadoop', 'group')},
            'oozie': {smoke_user, ('hadoop', 'group')},
        },
    }
    if zone.lower() != 'system':
        identities['users']['admin'] = ('admin', set())
    return identities


def cdp_identities(zone):
    """Identities needed for Cloudera Data Platform"""
    smoke_user = ('cloudera-scm', 'user')
    identities = {
        'groups': set(),  # Groups with no users in them.
        'users': {
            'accumulo': ('accumulo', set()),
            'anonymous': ('anonymous', set()),
            'apache': ('apache', set()),
            'atlas': ('atlas', {'hadoop', 'supergroup'}),
            'cloudera-scm': ('cloudera-scm', set()),
            'cmjobuser': ('cmjobuser', set()),
            'cruisecontrol': ('cruisecontrol', set()),
            'druid': ('druid', {'hadoop', 'supergroup'}),
            'flume': ('flume', set()),
            'hbase': ('hbase', {'hadoop', 'supergroup'}),
            'hdfs': ('hdfs', {'hadoop', 'supergroup'}),
            'hive': ('hive', set()),
            'HTTP': ('HTTP', {'hadoop', 'supergroup'}),
            'httpfs': ('httpfs', set()),
            'hue': ('hue', set()),
            'impala': ('impala', {'hive'}),
            'kafka': ('kafka', set()),
            'keytrustee': ('keytrustee', set()),
            'kms': ('kms', set()),
            'knox': ('knox', set()),
            'knoxui': ('knoxui', set()),
            'kudu': ('kudu', set()),
            'llama': ('llama', set()),
            'livy': ('livy', set()),
            'mapred': ('mapred', {'hadoop', 'supergroup'}),
            'oozie': ('oozie', set()),
            'phoenix': ('phoenix', set()),
            'ranger': ('ranger', {'hadoop', 'supergroup'}),
            'rangeradmin': ('rangeradmin', set()),
            'rangerlookup': ('rangerlookup', set()),
            'rangerraz': ('rangerraz', set()),
            'rangerrms': ('rangerrms', set()),
            'rangertagsync': ('rangertagsync', set()),
            'rangerusersync': ('rangerusersync', set()),
            'schemaregistry': ('schemaregistry', set()),
            'sentry': ('sentry', set()),
            'solr': ('solr', set()),
            'spark': ('spark', set()),
            'sqoop': ('sqoop', {'sqoop2'}),
            'sqoop2': ('sqoop2', {'sqoop'}),
            'streamsmsgmgr': ('streamsmsgmgr', set()),
            'tez': ('tez', set()),
            'superset': ('superset', set()),
            'yarn': ('yarn', {'hadoop', 'supergroup'}),
            'zeppelin': ('zeppelin', set()),
            'zookeeper': ('zookeeper', set()),
        },
        'proxy_users': {
            'flume': {smoke_user, ('hadoop', 'group')},
            'hive': {smoke_user, ('hadoop', 'group')},
            'hue': {smoke_user, ('hadoop', 'group')},
            'impala': {smoke_user, ('hadoop', 'group')},
            'mapred': {smoke_user, ('hadoop', 'group')},
            'oozie': {smoke_user, ('hadoop', 'group')},
            'phoenix': {smoke_user, ('hadoop', 'group')},
            'yarn': {smoke_user, ('hadoop', 'group')},
            'knox': {smoke_user, ('hadoop', 'group')},
            'hdfs': {smoke_user, ('hadoop', 'group')},
            'livy': {smoke_user, ('hadoop', 'group')},
            'HTTP': {smoke_user},
        },
    }
    if zone.lower() != 'system':
        identities['users']['admin'] = ('admin', set())
    return identities


def hdp_identities(zone):
    """Identities needed for Hortonworks Data Platform"""
    smoke_user = ('ambari-qa', 'user')
    identities = {
        'groups': set(),  # Groups with no users in them.
        'users': {
            'accumulo': ('accumulo', {'hadoop'}),
            'activity_analyzer': ('activity_analyzer', {'hadoop'}),
            'activity_explorer': ('activity_explorer', {'hadoop'}),
            'ambari-qa': ('ambari-qa', {'hadoop'}),
            'ambari-server': ('ambari-server', {'hadoop'}),
            'ams': ('ams', {'hadoop'}),
            'anonymous': ('anonymous', set()),
            'atlas': ('atlas', {'hadoop'}),
            'druid': ('druid', {'hadoop'}),
            'falcon': ('falcon', {'hadoop'}),
            'flume': ('flume', {'hadoop'}),
            'gpadmin': ('gpadmin', {'hadoop'}),
            'hadoopqa': ('hadoopqa', {'hadoop'}),
            'hbase': ('hbase', {'hadoop'}),
            'hcat': ('hcat', {'hadoop'}),
            'hdfs': ('hdfs', {'hadoop'}),
            'hive': ('hive', {'hadoop'}),
            'HTTP': ('HTTP', {'hadoop'}),
            'hue': ('hue', {'hadoop'}),
            'infra-solr': ('infra-solr', {'hadoop'}),
            'kafka': ('kafka', {'hadoop'}),
            'keyadmin': ('keyadmin', {'hadoop'}),
            'kms': ('kms', {'hadoop'}),
            'knox': ('knox', {'hadoop'}),
            'livy': ('livy', {'hadoop'}),
            'logsearch': ('logsearch', {'hadoop'}),
            'mahout': ('mahout', {'hadoop'}),
            'mapred': ('mapred', {'hadoop'}),
            'oozie': ('oozie', {'hadoop'}),
            'ranger': ('ranger', {'hadoop'}),
            'rangerlookup': ('rangerlookup', {'hadoop'}),
            'spark': ('spark', {'hadoop'}),
            'sqoop': ('sqoop', {'hadoop'}),
            'storm': ('storm', {'hadoop'}),
            'tez': ('tez', {'hadoop'}),
            'tracer': ('tracer', {'hadoop'}),
            'yarn': ('yarn', {'hadoop'}),
            'yarn-ats': ('yarn-ats', {'hadoop'}),
            'yarn-ats-hbase': ('yarn-ats-hbase', {'hadoop'}),
            'zeppelin': ('zeppelin', {'hadoop'}),
            'zookeeper': ('zookeeper', {'hadoop'}),
        },
        'proxy_users': {
            'ambari-server': {smoke_user},
            'flume': {smoke_user, ('hadoop', 'group')},
            'hbase': {smoke_user, ('hadoop', 'group')},
            'hcat': {smoke_user, ('hadoop', 'group')},
            'hive': {smoke_user, ('hadoop', 'group')},
            'HTTP': {smoke_user},
            'knox': {smoke_user},
            'livy': {smoke_user, ('hadoop', 'group')},
            'oozie': {smoke_user, ('hadoop', 'group')},
            'yarn': {smoke_user, ('hadoop', 'group')},
        },
    }
    if zone.lower() != 'system':
        identities['users']['admin'] = ('admin', set())
    return identities



import pytest

import isilon_hadoop_tools.identities


@pytest.mark.parametrize('zone', ['System', 'notSystem'])
@pytest.mark.parametrize(
    'identities',
    [
        isilon_hadoop_tools.identities.cdh_identities,
        isilon_hadoop_tools.identities.cdp_identities,
        isilon_hadoop_tools.identities.hdp_identities,
    ],
)
def test_log_identities(identities, zone):
    """Verify that log_identities returns None."""
    assert isilon_hadoop_tools.identities.log_identities(identities(zone)) is None



"""Classes for Interacting with OneFS"""

# pylint: disable=too-many-lines

from __future__ import absolute_import
from __future__ import unicode_literals

from datetime import date, datetime
from enum import Enum
import json
import logging
import posixpath
import socket
import struct
import time
try:
    from urllib.parse import urlparse, urlunparse  # Python 3
except ImportError:
    from urlparse import urlparse, urlunparse  # Python 2
import urllib3

from future.utils import raise_from
import requests

from isilon_hadoop_tools import IsilonHadoopToolError


__all__ = [
    # Constants / Enums
    'ONEFS_RELEASES',
    'OneFSFeature',

    # Decorators
    'accesses_onefs',

    # Exceptions
    'APIError',
    'ExpiredLicenseError',
    'MalformedAPIError',
    'MissingLicenseError',
    'MissingZoneError',
    'MixedModeError',
    'NonSDKAPIError',
    'OneFSError',
    'OneFSConnectionError',
    'OneFSCertificateError',
    'OneFSValueError',
    'UndecodableAPIError',
    'UndeterminableVersion',
    'UnsupportedOperation',
    'UnsupportedVersion',

    # Functions
    'sdk_for_revision',

    # Objects
    'Client',
]

LOGGER = logging.getLogger(__name__)
ONEFS_RELEASES = {
    '7.2.0.0': 0x70200500000000A,
    '8.0.0.0': 0x800005000000025,
    '8.0.0.4': 0x800005000400035,
    '8.0.1.0': 0x800015000000007,
    '8.0.1.1': 0x800015000100070,
    '8.1.0.0': 0x80100500000000B,
    '8.1.1.0': 0x8010150000000D4,
    '8.1.2.0': 0x801025000000010,
    '8.1.3.0': 0x80103500000000D,
    '8.2.0.0': 0x80200500000000B,
    '8.2.1.0': 0x802015000000004,
    '8.2.2.0': 0x802025000000007,
    '8.2.3.0': 0x802035000000000,
}


class OneFSFeature(Enum):

    """OneFS Feature Flags for Use with Client.feature_is_supported"""

    # These values come from sys/sys/isi_upgrade_api_featuremap.h.
    # For example,
    #  ISI_UPGRADE_API_FEATURE_VERSION(PIPE_UAPI_OVERRIDES, 8, 1, 3, 0)
    # translates to
    #  PIPE_UAPI_OVERRIDES = (0x8010300, 0)

    _GEN = {
        'INIT': 0x0000000,
        'JAWS': 0x7010100,
        'MOBY': 0x7020000,
        'ORCA': 0x7020100,
        'RIP0': 0x7030000,
        'RIP1': 0x7030100,
        'RIPT': 0x8000000,
        'HAPI': 0x8000100,
        'FRTR': 0x8010000,
        'NJMA': 0x8010100,
        'KANA': 0x8010200,
        'NDUU': 0x8010300,
        'PIPE': 0x8020000,
        'ERA1': 0x9000100,
    }

    # pylint: disable=invalid-name
    FOREVER = (_GEN['INIT'], 0)

    JAWS_RU = (_GEN['JAWS'], 0)

    MOBY_PROTECTION = (_GEN['MOBY'], 0)
    MOBY_SNAPDELETE = (_GEN['MOBY'], 1)
    MOBY_RU = (_GEN['MOBY'], 2)
    MOBY_UNFS = (_GEN['MOBY'], 3)
    MOBY_AUTH_UPGRADE = (_GEN['MOBY'], 4)

    ORCA_RU = (_GEN['ORCA'], 0)
    RIPT_CONSISTENT_HASH = (_GEN['ORCA'], 1)

    RIPT_RBM_VERSIONING = (_GEN['RIP0'], 0)

    BATCH_ERROR_DSR = (_GEN['RIP1'], 1)
    RIPTIDE_MEDIASCAN = (_GEN['RIP1'], 2)
    RIPT_8K_INODES = (_GEN['RIP1'], 3)

    RIPT_DEDUPE = (_GEN['RIPT'], 0)
    RIPTIDE_TRUNCATE = (_GEN['RIPT'], 1)
    RIPTIDE_CHANGELISTCREATE = (_GEN['RIPT'], 2)
    RIPT_GMP_SERVICES = (_GEN['RIPT'], 3)
    RIPT_NLM = (_GEN['RIPT'], 4)
    RIPTIDE_FSA = (_GEN['RIPT'], 5)
    RIPT_SMARTPOOLS = (_GEN['RIPT'], 6)
    RIPT_AUTH_UPGRADE = (_GEN['RIPT'], 7)
    RIPT_CELOG_UPGRADE = (_GEN['RIPT'], 8)

    HALFPIPE_PARTITIONED_PERFORMANCE = (_GEN['HAPI'], 0)
    HP_JE = (_GEN['HAPI'], 1)
    HP_WORM = (_GEN['HAPI'], 2)
    HP_PROXY = (_GEN['HAPI'], 3)
    HALFPIPE_CONTAINERS = (_GEN['HAPI'], 4)
    HP_NEEDS_NDU_FLAG = (_GEN['HAPI'], 5)
    HP_RANGER = (_GEN['HAPI'], 6)
    HP_AMBARI_METRICS = (_GEN['HAPI'], 7)
    HP_DATANODE_WIRE_ENCRYPTION = (_GEN['HAPI'], 8)

    FT_SMARTPOOLS = (_GEN['FRTR'], 0)
    FRT_MIRRORED_JOURNAL = (_GEN['FRTR'], 1)
    FRT_LIN_SUPER_DRIVE_QUORUM = (_GEN['FRTR'], 2)
    FREIGHT_TRAINS_LAYOUT = (_GEN['FRTR'], 3)
    FT_ESRS = (_GEN['FRTR'], 4)
    FRT_COMPRESSED_INODES = (_GEN['FRTR'], 5)
    FTR_LICENSE_MIGRATION = (_GEN['FRTR'], 6)

    PIPE_ITER_MARK = (_GEN['NJMA'], 0)
    NIIJIMA_CPOOL_GOOGLE_XML = (_GEN['NJMA'], 1)
    NJMA_HDFS_INOTIFY = (_GEN['NJMA'], 2)
    NJMA_HDFS_FSIMAGE = (_GEN['NJMA'], 3)
    NIIJIMA_CLUSTER_TIME = (_GEN['NJMA'], 4)
    NIIJIMA_SMB = (_GEN['NJMA'], 5)
    NIIJIMA_ESRS = (_GEN['NJMA'], 6)

    KANA_HDFS_REF_BY_INODE = (_GEN['KANA'], 0)
    KANA_WEBHDFS_DELEGATION_TOKENS = (_GEN['KANA'], 1)

    PIPE_UAPI_OVERRIDES = (_GEN['NDUU'], 0)

    PIPE_AUTH_AWS_V4 = (_GEN['PIPE'], 0)
    PIPE_HANGDUMP = (_GEN['PIPE'], 1)
    PIPE_GMP_CFG_GEN = (_GEN['PIPE'], 2)
    PIPE_EXT_GROUP = (_GEN['PIPE'], 3)
    PIPE_EXT_GROUP_MSG = (_GEN['PIPE'], 4)
    PIPE_JE = (_GEN['PIPE'], 5)
    PIPE_IFS_DOMAINS = (_GEN['PIPE'], 6)
    PIPE_CPOOL_SECURE_KEY = (_GEN['PIPE'], 7)
    PIPE_CPOOL_C2S = (_GEN['PIPE'], 8)
    PIPE_ISI_CERTS = (_GEN['PIPE'], 9)
    PIPE_NDMP = (_GEN['PIPE'], 10)
    PIPE_ZONED_ROLES = (_GEN['PIPE'], 11)
    FT_JE_ZOMBIE = (_GEN['PIPE'], 12)
    PIPE_CPOOL_GOOGLE_XML = (_GEN['PIPE'], 13)
    PIPE_SIQ = (_GEN['PIPE'], 14)
    PIPE_QUOTAS_MS = (_GEN['PIPE'], 15)
    PIPE_QUOTA_USER_CONTAINERS = (_GEN['PIPE'], 16)
    PIPE_TREEDELETE = (_GEN['PIPE'], 17)
    PIPE_DOMAIN_SNAPSHOTS = (_GEN['PIPE'], 18)
    PIPE_QUOTA_DDQ = (_GEN['PIPE'], 19)
    PIPE_ARRAYD = (_GEN['PIPE'], 20)
    PIPE_FLEXNET_V4 = (_GEN['PIPE'], 21)
    PIPE_ISI_DAEMON_IPV6 = (_GEN['PIPE'], 22)
    PIPE_JE_PREP = (_GEN['PIPE'], 23)
    PIPE_IFS_LFN = (_GEN['PIPE'], 24)
    PIPE_SNAP_SCHED_TARDIS = (_GEN['PIPE'], 25)
    PIPE_DRIVE_INTEROP = (_GEN['PIPE'], 26)
    PIPE_READ_BLOCKS = (_GEN['PIPE'], 27)
    PIPE_IFS_BCM = (_GEN['PIPE'], 28)
    PIPE_EXT_GRP_SRO = (_GEN['PIPE'], 29)
    PIPE_HDFS_EXTATTR = (_GEN['PIPE'], 30)
    PIPE_CP_2_0 = (_GEN['PIPE'], 31)
    PIPE_FILEPOLICY = (_GEN['PIPE'], 32)
    PIPE_COAL_SUSP_AGGR = (_GEN['PIPE'], 34)
    PIPE_SMARTCONNECT_DNS = (_GEN['PIPE'], 35)
    PIPE_PDM_ENC_INATTR = (_GEN['PIPE'], 36)
    PIPE_NDMP_REDIRECTOR = (_GEN['PIPE'], 38)
    PIPE_ISI_CBIND_D = (_GEN['PIPE'], 39)
    PIPE_SPARSE_PUNCH = (_GEN['PIPE'], 41)
    PIPE_SSH_CONFIG = (_GEN['PIPE'], 42)
    PIPE_AUDIT_EVENTS = (_GEN['PIPE'], 43)
    PIPE_PURPOSEDB = (_GEN['PIPE'], 45)
    PIPE_JE_TREEWALK = (_GEN['PIPE'], 47)

    ERA1_HDFS_TDE = (_GEN['ERA1'], 1)
    ERA1_QUOTA_APPLOGICAL = (_GEN['ERA1'], 4)
    ERA1_IDI_VERIFY_SNAPID = (_GEN['ERA1'], 6)
    ERA1_CPOOL_ALIYUN = (_GEN['ERA1'], 7)
    ERA1_STF_DUMMY_LINS = (_GEN['ERA1'], 8)
    ERA1_PDM_COLLECT = (_GEN['ERA1'], 13)
    ERA1_MCP_MLIST = (_GEN['ERA1'], 14)
    ERA1_NFS_SCHED_CONFIG = (_GEN['ERA1'], 16)
    ERA1_ADS_VOPS = (_GEN['ERA1'], 17)
    ERA1_GMP_SERVICE_LSASS = (_GEN['ERA1'], 18)
    ERA1_SINLIN_LOCK_ORDER = (_GEN['ERA1'], 20)
    ERA1_LIN_MASTER_FLAGS = (_GEN['ERA1'], 23)
    ERA1_REMOTE_SYSCTL_OBJECT = (_GEN['ERA1'], 25)
    ERA1_LIN_BUCKET_LOCK = (_GEN['ERA1'], 27)
    ERA1_PDM_SNAPGOV_RENAME = (_GEN['ERA1'], 34)
    # pylint: enable=invalid-name


class OneFSError(IsilonHadoopToolError):
    """All Exceptions emitted from this module inherit from this Exception."""


class OneFSConnectionError(OneFSError):
    """
    This Exception is raised when a client cannot connect to OneFS
    (e.g. due to socket.gaierror or urllib3.exceptions.MaxRetryError).
    """


class OneFSCertificateError(OneFSConnectionError):
    """This exception occurs when a client cannot connect due to an invalid HTTPS certificate."""


class NonSDKAPIError(OneFSError):
    """This exception is raised when interacting with OneFS APIs without the SDK fails."""


class _BaseAPIError(OneFSError):

    def __init__(self, exc):
        super(_BaseAPIError, self).__init__(str(exc))
        self.exc = exc


class MalformedAPIError(_BaseAPIError):
    """This exception wraps an Isilon SDK ApiException that does not have a valid JSON body."""


class UndecodableAPIError(MalformedAPIError):
    """This exception wraps an Isilon SDK ApiException that does not have a JSON-decodable body."""


class APIError(_BaseAPIError):

    """This exception wraps an Isilon SDK ApiException."""

    # pylint: disable=invalid-name
    gid_already_exists_error_format = "Group already exists with gid '{0}'"
    group_already_exists_error_format = "Group '{0}' already exists"
    group_not_found_error_format = "Failed to find group for 'GROUP:{0}': No such group"
    group_unresolvable_error_format = "Could not resolve group {0}"
    license_expired_error_format = (
        # Note: Subscriptions did not exist prior to Freight Trains,
        # so old code assumes that only evaluations have expiration dates.
        "The evaluation license key for {0} has expired."
        " Please contact your Isilon representative."
    )
    license_missing_error_format = (
        "The {0} application is not currently installed."
        " Please contact your Isilon account team for"
        " more information on evaluating and purchasing {0}."
    )
    proxy_user_already_exists_error_format = "Proxyuser '{0}' already exists"
    try_again_error_format = "OneFS API is temporarily unavailable. Try your request again."
    uid_already_exists_error_format = "User already exists with uid '{0}'"
    user_already_exists_error_format = "User '{0}' already exists"
    user_already_in_group_error_format = \
        "Failed to add member UID:{0} to group GROUP:{1}: User is already in local group"
    user_not_found_error_format = "Failed to find user for 'USER:{0}': No such user"
    user_unresolvable_error_format = "Could not resolve user {0}"
    zone_not_found_error_format = 'Access Zone "{0}" not found.'
    dir_path_already_exists_error_format = \
        'Unable to create directory as requested -- container already exists'
    # pylint: enable=invalid-name

    def __str__(self):
        try:
            return '\n'.join(error['message'] for error in self.errors()) or str(self.exc)
        except MalformedAPIError as exc:
            return str(exc)

    def errors(self):
        """Get errors listed in the exception."""
        try:
            json_body = json.loads(self.exc.body)
        except (
                TypeError,   # self.exc.body is not a str.
                ValueError,  # self.exc.body is not JSON.
        ) as exc:
            raise_from(UndecodableAPIError(self.exc), exc)
        try:
            for error in json_body['errors']:
                # Raise a KeyError if 'message' is not in error:
                error['message']  # pylint: disable=pointless-statement
        except (
                KeyError,   # 'errors' or 'message' is not in json_body or error, respectively.
                TypeError,  # json_body['errors'] is not iterable.
        ) as exc:
            raise_from(MalformedAPIError(self.exc), exc)
        return json_body['errors']

    def filtered_errors(self, filter_func):
        """Arbitrarily filter errors in the exception."""
        for error in self.errors():
            if filter_func(error):
                yield error

    def gid_already_exists_error(self, gid):
        """Returns True if the exception contains a GID already exists error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.gid_already_exists_error_format.format(gid),
            )
        )

    def group_already_exists_error(self, group_name):
        """Returns True if the exception contains a group already exists error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.group_already_exists_error_format.format(
                    group_name,
                ),
            )
        )

    def group_not_found_error(self, group_name):
        """Returns True if the exception contains a group not found error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.group_not_found_error_format.format(
                    group_name,
                ),
            )
        )

    def group_unresolvable_error(self, group_name):
        """Returns True if the exception contains an unresolvable group error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.group_unresolvable_error_format.format(
                    group_name,
                ),
            )
        )

    def license_expired_error(self, license_name):
        """Returns True if the exception contains an expired license error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.license_expired_error_format.format(
                    license_name,
                ),
            )
        )

    def license_missing_error(self, license_name):
        """Returns True if the exception contains a missing license error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.license_missing_error_format.format(
                    license_name,
                ),
            )
        )

    def proxy_user_already_exists_error(self, proxy_user_name):
        """Returns True if the exception contains a proxy user already exists error."""
        return any(
            self.filtered_errors(
                lambda error:
                error['message'] == self.proxy_user_already_exists_error_format.format(
                    proxy_user_name,
                ),
            )
        )

    def try_again_error(self):
        """Returns True if the exception indicated PAPI is temporarily unavailable."""
        return any(
            self.filtered_errors(
                lambda error:
                error['message'] == self.try_again_error_format,
            )
        )

    def uid_already_exists_error(self, uid):
        """Returns True if the exception contains a UID already exists error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.uid_already_exists_error_format.format(uid),
            )
        )

    def user_already_exists_error(self, user_name):
        """Returns True if the exception contains a user already exists error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.user_already_exists_error_format.format(
                    user_name,
                ),
            )
        )

    def user_already_in_group_error(self, uid, group_name):
        """Returns True if the exception contains a user already in group error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.user_already_in_group_error_format.format(
                    uid,
                    group_name,
                ),
            )
        )

    def user_not_found_error(self, user_name):
        """Returns True if the exception contains a user not found error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.user_not_found_error_format.format(
                    user_name,
                ),
            )
        )

    def user_unresolvable_error(self, user_name):
        """Returns True if the exception contains an unresolvable user error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.user_unresolvable_error_format.format(
                    user_name,
                ),
            )
        )

    def zone_not_found_error(self, zone_name):
        """Returns True if the exception contains a zone not found error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.zone_not_found_error_format.format(
                    zone_name,
                ),
            )
        )

    def dir_path_already_exists_error(self):
        """Returns True if the exception contains a directory path already exist error."""
        return any(
            self.filtered_errors(
                lambda error: error['message'] == self.dir_path_already_exists_error_format,
            )
        )


class MissingLicenseError(OneFSError):
    """This Exception is raised when a license that is expected to exist cannot be found."""


class ExpiredLicenseError(MissingLicenseError):
    """This Exception is raised when a license has expired."""


class MissingZoneError(OneFSError):
    """This Exception is raised when a zone that is expected to exist cannot be found."""


class MixedModeError(OneFSError):
    """
    This Exception is raised when an operation cannot succeed due to
    the cluster containing nodes running different versions of OneFS.
    """


class UndeterminableVersion(OneFSError):
    """
    This Exception is raised when attempting to use this
    module with a version of OneFS that cannot be determined (usually < 8.0.0.0).
    """


class UnsupportedVersion(OneFSError):
    """
    This Exception is raised when attempting to use this
    module with an unsupported version of OneFS.
    """


class UnsupportedOperation(OneFSError):
    """
    This Exception is raised when attempting to conduct an unsupported
    operation with an specific version of OneFS.
    """


def sdk_for_revision(revision, strict=False):
    """Get the SDK that is intended to work with a given OneFS revision."""
    # pylint: disable=too-many-return-statements,import-outside-toplevel
    if ONEFS_RELEASES['7.2.0.0'] <= revision < ONEFS_RELEASES['8.0.0.0']:
        import isi_sdk_7_2
        return isi_sdk_7_2
    if ONEFS_RELEASES['8.0.0.0'] <= revision < ONEFS_RELEASES['8.0.1.0']:
        import isi_sdk_8_0
        return isi_sdk_8_0
    if ONEFS_RELEASES['8.0.1.0'] <= revision < ONEFS_RELEASES['8.1.0.0']:
        import isi_sdk_8_0_1
        return isi_sdk_8_0_1
    if ONEFS_RELEASES['8.1.0.0'] <= revision < ONEFS_RELEASES['8.1.1.0']:
        import isi_sdk_8_1_0
        return isi_sdk_8_1_0
    if ONEFS_RELEASES['8.1.1.0'] <= revision < ONEFS_RELEASES['8.2.0.0']:
        import isi_sdk_8_1_1
        return isi_sdk_8_1_1
    if ONEFS_RELEASES['8.2.0.0'] <= revision < ONEFS_RELEASES['8.2.1.0']:
        import isi_sdk_8_2_0
        return isi_sdk_8_2_0
    if ONEFS_RELEASES['8.2.1.0'] <= revision < ONEFS_RELEASES['8.2.2.0']:
        import isi_sdk_8_2_1
        return isi_sdk_8_2_1
    if ONEFS_RELEASES['8.2.2.0'] <= revision < ONEFS_RELEASES['8.2.3.0']:
        import isi_sdk_8_2_2
        return isi_sdk_8_2_2
    # At this point, either the cluster is too old or too new;
    # however, new clusters still support old SDKs,
    # so, unless the caller asks to fail here, we'll fall back to the newest supported SDK.
    if strict:
        raise UnsupportedVersion('There is no SDK for OneFS revision {0}!'.format(hex(revision)))
    import isi_sdk_8_2_2
    return isi_sdk_8_2_2  # The latest SDK available.


def accesses_onefs(func):
    """Decorate a Client method that makes an SDK call directly."""
    def _decorated(self, *args, **kwargs):
        while True:
            try:
                return func(self, *args, **kwargs)
            except urllib3.exceptions.MaxRetryError as exc:
                if isinstance(exc.reason, urllib3.exceptions.SSLError):
                    # https://github.com/Isilon/isilon_sdk_python/issues/14
                    raise_from(OneFSCertificateError, exc)
                raise_from(OneFSConnectionError, exc)
            except self._sdk.rest.ApiException as exc:  # pylint: disable=protected-access
                if all([
                        # https://github.com/PyCQA/pylint/issues/2841
                        not exc.body,  # pylint: disable=no-member
                        'CERTIFICATE_VERIFY_FAILED' in (
                            exc.reason or ''  # pylint: disable=no-member
                        ),
                ]):
                    raise_from(OneFSCertificateError, exc)
                wrapped_exc = APIError(exc)
                if not wrapped_exc.try_again_error():
                    raise_from(wrapped_exc, exc)
                time.sleep(2)
                LOGGER.info(wrapped_exc.try_again_error_format)
    return _decorated


class OneFSValueError(OneFSError, ValueError):
    """
    This exception is raised by this module instead
    of a ValueError (but has the same meaning).
    """


def _license_is_active(license_):
    return license_.status.lower() in ['activated', 'evaluation', 'licensed']


class BaseClient(object):  # pylint: disable=too-many-public-methods,too-many-instance-attributes

    """Interact with OneFS."""

    def __init__(self, address, username, password, default_zone='System', verify_ssl=True):

        # Set attributes without setters first.
        self.default_zone = default_zone
        # We don't know what version we are pointed at yet, but we have to start somewhere.
        # Riptide was the first SDK to support ClusterApi().get_cluster_version.
        self._sdk = sdk_for_revision(ONEFS_RELEASES['8.0.0.0'])
        # Attributes with setters (see below) depend on having a Configuration object to manipulate.
        self._configuration = self._sdk.Configuration()
        self._address = None  # This will truly be set last.

        # Set attributes with setters last.
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        # Set the address last so the rest of the configuration is in place for making requests
        # (which will be needed for checking the cluster version).
        self.address = address

    @property
    def _api_client(self):
        return self._sdk.ApiClient(self._configuration)

    @accesses_onefs
    def _groups(self, zone=None):
        return self._sdk.AuthApi(self._api_client).list_auth_groups(
            zone=zone or self.default_zone,
        ).groups

    @accesses_onefs
    def _keytab_entries(self, provider):
        providers_krb5 = self._sdk.AuthApi(self._api_client).get_providers_krb5_by_id(provider)
        return providers_krb5.krb5[0].keytab_entries

    @accesses_onefs
    def _license(self, name):
        return self._sdk.LicenseApi(self._api_client).get_license_license(name)

    @accesses_onefs
    def _pools(self, *args, **kwargs):
        return self._sdk.NetworkApi(self._api_client).get_network_pools(*args, **kwargs).pools

    @accesses_onefs
    def _realms(self):
        return self._sdk.AuthApi(self._api_client).list_settings_krb5_realms().realm

    def _refresh_sdk(self):
        try:
            self._revision = self.revision()  # pylint: disable=attribute-defined-outside-init
        except AttributeError as exc:
            raise_from(UndeterminableVersion, exc)
        self._sdk = sdk_for_revision(self._revision)

    @accesses_onefs
    def _upgrade_cluster(self):
        return self._sdk.UpgradeApi(self._api_client).get_upgrade_cluster()

    @accesses_onefs
    def _version(self):
        return self._sdk.ClusterApi(self._api_client).get_cluster_version()

    def _zone(self, name):
        for zone in self._zones():
            # Zone names are NOT case-sensitive.
            if zone.name.lower() == name.lower():
                return zone
        raise MissingZoneError(name)

    def _zone_real_path(self, path, zone=None):
        return posixpath.join(
            self.zone_settings(zone=zone or self.default_zone)['path'],
            path.lstrip(posixpath.sep),
        )

    @accesses_onefs
    def _zones(self):
        return self._sdk.ZonesApi(self._api_client).list_zones().zones

    @accesses_onefs
    def acl_settings(self):
        """Get global ACL settings."""
        acl_settings = self._sdk.AuthApi(self._api_client).get_settings_acls().acl_policy_settings
        return {
            'access': acl_settings.access,
            'calcmode': acl_settings.calcmode,
            'calcmode_group': acl_settings.calcmode_group,
            'calcmode_owner': acl_settings.calcmode_owner,
            'chmod': acl_settings.chmod,
            'chmod_007': acl_settings.chmod_007,
            'chmod_inheritable': acl_settings.chmod_inheritable,
            'chown': acl_settings.chown,
            'create_over_smb': acl_settings.create_over_smb,
            'dos_attr': acl_settings.dos_attr,
            'group_owner_inheritance': acl_settings.group_owner_inheritance,
            'rwx': acl_settings.rwx,
            'synthetic_denies': acl_settings.synthetic_denies,
            'utimes': acl_settings.utimes,
        }

    @accesses_onefs
    def add_user_to_group(self, user_name, group_name, zone=None):
        """Add a user to a group."""
        group_member_cls = (
            self._sdk.GroupMember
            if self._revision < ONEFS_RELEASES['8.0.1.0'] else
            self._sdk.AuthAccessAccessItemFileGroup
        )
        try:
            self._sdk.AuthGroupsApi(self._api_client).create_group_member(
                group_member_cls(
                    type='user',
                    name=user_name,
                ),
                group_name,
                zone=zone or self.default_zone,
            )
        except ValueError as exc:
            # https://bugs.west.isilon.com/show_bug.cgi?id=231922
            assert all([
                str(exc) == 'Invalid value for `id`, must not be `None`',
                user_name in [
                    member.name
                    for member in self._sdk.AuthGroupsApi(self._api_client).list_group_members(
                        group_name,
                        zone=zone or self.default_zone,
                    ).members
                ],
            ])

    @property
    def address(self):
        """Get the address to connect to OneFS at."""
        # self._address may be None if self.host was set directly.
        return self._address or urlparse(self.host).hostname

    @address.setter
    def address(self, address):
        """
        Set the address to connect to OneFS at.
        If the address is a name, it will be resolved first to avoid config propagation problems.
        To avoid that, set host instead.
        """
        try:
            # If address is a SmartConnect name, making calls too fast can result in errors
            # due to changes not propagating fast enough across a cluster.
            # This problem gets worse on larger clusters.
            # So, we will choose 1 node to connect to and use that.
            netloc = socket.gethostbyname(address)
        except socket.gaierror as exc:
            raise_from(OneFSConnectionError, exc)
        if ':' in netloc:
            netloc = '[{ipv6}]'.format(ipv6=netloc)

        # Keep every part of self.host, except the hostname/address.
        parsed = urlparse(self.host)
        if parsed.port is not None:
            netloc += ':' + str(parsed.port)
        self.host = urlunparse(parsed._replace(netloc=netloc))

        # Setting self.host unsets self._address:
        self._address = address

    def check_license(self, name):
        """Check for a license on OneFS and raise a MissingLicenseError if it doesn't exist."""
        [license_] = self._license(name).licenses
        if not _license_is_active(license_):
            if license_.expiration and \
                    datetime.strptime(license_.expiration, '%Y-%m-%d').date() < date.today():
                raise ExpiredLicenseError(name)
            raise MissingLicenseError(name)

    def check_zone(self, name):
        """Check for a zone on OneFS and raise a MissingZoneError if it doesn't exist."""
        if not self.has_zone(name):
            raise MissingZoneError(name)

    @accesses_onefs
    def chmod(self, path, mode, zone=None):
        """Change the (integer) mode of a (zone-root-relative) path."""
        real_path = self._zone_real_path(path, zone=zone or self.default_zone)
        self._sdk.NamespaceApi(self._api_client).set_acl(
            namespace_path=real_path.lstrip(posixpath.sep),
            acl=True,
            namespace_acl=self._sdk.NamespaceAcl(
                authoritative='mode',
                mode='{:o}'.format(mode),
            ),
        )

    @accesses_onefs
    def chown(self, path, owner=None, group=None, zone=None):
        """Change the owning user and/or group of a (zone-root-relative) path."""
        real_path = self._zone_real_path(path, zone=zone)
        ns_acl_kwargs = {'authoritative': 'mode'}
        if owner is not None:
            # Get the UID of the owner to avoid name resolution problems across zones
            # (e.g. using the System zone to configure a different zone).
            ns_acl_kwargs['owner'] = self._sdk.MemberObject(
                type='UID',
                id='UID:{uid}'.format(
                    uid=owner if isinstance(owner, int) else self.uid_of_user(owner, zone=zone),
                ),
            )
        if group is not None:
            # Get the GID of the group to avoid name resolution problems across zones
            # (e.g. using the System zone to configure a different zone).
            ns_acl_kwargs['group'] = self._sdk.MemberObject(
                type='GID',
                id='GID:{gid}'.format(
                    gid=group if isinstance(group, int) else self.gid_of_group(group, zone=zone),
                ),
            )
        self._sdk.NamespaceApi(self._api_client).set_acl(
            namespace_path=real_path.lstrip(posixpath.sep),
            acl=True,
            namespace_acl=self._sdk.NamespaceAcl(**ns_acl_kwargs),
        )

    @accesses_onefs
    def create_auth_provider(self, realm, user, password):
        """Create a Kerberos auth provider."""
        self._sdk.AuthApi(self._api_client).create_providers_krb5_item(
            self._sdk.ProvidersKrb5Item(
                realm=realm,
                user=user,
                password=password,
            ),
        )

    @accesses_onefs
    def create_group(self, name, gid=None, zone=None):
        """Create a group."""
        self._sdk.AuthApi(self._api_client).create_auth_group(
            self._sdk.AuthGroupCreateParams(
                name=name,
                gid=gid,
            ),
            zone=zone or self.default_zone,
        )

    @accesses_onefs
    def create_hdfs_proxy_user(self, name, members=None, zone=None):
        """Create an HDFS proxy user."""
        if members is not None:
            group_member_cls = (
                self._sdk.GroupMember
                if self._revision < ONEFS_RELEASES['8.0.1.0'] else
                self._sdk.AuthAccessAccessItemFileGroup
            )
            members = [
                group_member_cls(
                    name=member_name,
                    type=member_type,
                )
                for member_name, member_type in members
            ]
        self._sdk.ProtocolsApi(self._api_client).create_hdfs_proxyuser(
            self._sdk.HdfsProxyuserCreateParams(name=name, members=members),
            zone=zone or self.default_zone,
        )

    @accesses_onefs
    def create_realm(self, name, admin_server, kdcs):
        """Create a realm configuration on OneFS."""
        try:
            self._sdk.AuthApi(self._api_client).create_settings_krb5_realm(
                self._sdk.SettingsKrb5RealmCreateParams(
                    realm=name,
                    admin_server=admin_server,
                    kdc=kdcs,
                ),
            )
        except ValueError as exc:
            # https://bugs.west.isilon.com/show_bug.cgi?id=231054
            auth_api = self._sdk.AuthApi(self._api_client)
            assert all([
                str(exc) == 'Invalid value for `id`, must not be `None`',
                name in [krb5.realm for krb5 in auth_api.list_settings_krb5_realms().realm],
            ])

    @accesses_onefs
    def create_spn(self, spn, realm, user, password):
        """Create an SPN in a Kerberos realm."""
        providers_krb5_item = self._sdk.ProvidersKrb5Item(
            realm=realm,
            user=user,
            password=password,
        )
        keytab_entry = self._sdk.ProvidersKrb5IdParamsKeytabEntry()
        keytab_entry.spn = spn
        providers_krb5_item.keytab_entries = [keytab_entry]
        self._sdk.AuthApi(self._api_client).create_providers_krb5_item(providers_krb5_item)

    @accesses_onefs
    def create_user(self, name, primary_group_name, uid=None, zone=None, enabled=None):
        """Create a user."""
        group_member_cls = (
            self._sdk.GroupMember
            if self._revision < ONEFS_RELEASES['8.0.1.0'] else
            self._sdk.AuthAccessAccessItemFileGroup
        )
        self._sdk.AuthApi(self._api_client).create_auth_user(
            self._sdk.AuthUserCreateParams(
                name=name,
                enabled=enabled,
                primary_group=group_member_cls(
                    type='group',
                    name=primary_group_name,
                ),
                uid=uid,
            ),
            zone=zone or self.default_zone,
        )

    @accesses_onefs
    def delete_auth_provider(self, name):
        """Delete a Kerberos auth provider."""
        self._sdk.AuthApi(self._api_client).delete_providers_krb5_by_id(name)

    @accesses_onefs
    def delete_group(self, name, zone=None):
        """Delete a group."""
        self._sdk.AuthApi(self._api_client).delete_auth_group(
            name,
            zone=zone or self.default_zone,
        )

    @accesses_onefs
    def delete_hdfs_proxy_user(
            self,
            name,
            zone=None,
    ):
        """Delete an HDFS proxy user."""
        self._sdk.ProtocolsApi(self._api_client).delete_hdfs_proxyuser(
            name,
            zone=zone or self.default_zone,
        )

    @accesses_onefs
    def delete_realm(self, name):
        """Delete a Kerberos realm configuration."""
        self._sdk.AuthApi(self._api_client).delete_settings_krb5_realm(name)

    @accesses_onefs
    def delete_spn(self, spn, provider):
        """Delete a Kerberos SPN."""
        self._sdk.AuthApi(self._api_client).update_providers_krb5_by_id(
            self._sdk.ProvidersKrb5IdParams(
                keytab_entries=[
                    keytab_entry
                    for keytab_entry in self._keytab_entries(provider=provider)
                    if keytab_entry.spn not in [spn, spn + '@' + provider]
                ]
            ),
            provider,
        )

    @accesses_onefs
    def delete_user(self, name, zone=None):
        """Delete a user."""
        self._sdk.AuthApi(self._api_client).delete_auth_user(
            name,
            zone=zone or self.default_zone,
        )

    def feature_is_supported(self, feature):
        """Determine if a given OneFSFeature is supported."""

        feature_gen, feature_bit = feature.value

        upgrade_cluster = self._upgrade_cluster()
        try:
            committed_features = upgrade_cluster.committed_features
        except AttributeError as exc:
            raise_from(
                UnsupportedOperation(
                    'OneFS 8.2.0 or later is required for feature flag support.',
                ),
                exc,
            )

        entries_for_gen = [
            entry.bits for entry in committed_features.gen_bits
            if entry.gen == feature_gen
        ]
        if not entries_for_gen:
            return bool(feature_gen <= committed_features.default_gen)

        return any(
            feature_bit == (i * 64) + offset  # Each entry can have up to 64 offsets.
            for i, offsets in enumerate(entries_for_gen)
            for offset in offsets
        )

    @accesses_onefs
    def flush_auth_cache(self, zone=None):
        """Flush the Security Objects Cache."""
        if self._revision < ONEFS_RELEASES['8.0.1.0']:
            _zone = zone or self.default_zone
            if _zone and _zone.lower() != 'system':
                raise UnsupportedOperation(
                    'The auth cache can only be flushed on the System zone before OneFS 8.0.1.',
                )
            response = requests.delete(
                url=self.host + '/platform/3/auth/users',
                verify=self.verify_ssl,
                auth=(self.username, self.password),
                params={'cached': True},
            )
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as exc:
                raise_from(NonSDKAPIError('The auth cache could not be flushed.'), exc)
            else:
                assert bool(
                    response.status_code == requests.codes.no_content,  # pylint: disable=no-member
                )
        else:
            try:
                self._sdk.AuthApi(self._api_client).create_auth_cache_item(
                    auth_cache_item=self._sdk.AuthCacheItem(all='all'),
                    zone=zone or self.default_zone,
                )
            except ValueError as exc:
                # https://bugs.west.isilon.com/show_bug.cgi?id=232142
                assert str(exc) == 'Invalid value for `id`, must not be `None`'

    @accesses_onefs
    def gid_of_group(self, group_name, zone=None):
        """Get the GID of a group."""
        auth_groups = self._sdk.AuthApi(self._api_client).get_auth_group(
            group_name,
            zone=zone or self.default_zone,
        )
        assert len(auth_groups.groups) == 1, 'Do you have duplicate groups (e.g. local and LDAP)?'
        return int(auth_groups.groups[0].gid.id.split(':')[1])

    def groups(self, zone=None):
        """Get the auth groups OneFS knows about."""
        for group in self._groups(zone=zone or self.default_zone):
            yield group.name

    def has_license(self, name):
        """Check for a OneFS license on OneFS."""
        return any(_license_is_active(license) for license in self._license(name).licenses)

    def has_zone(self, name):
        """Check for a zone on OneFS."""
        return self._zone(name) is not None

    @accesses_onefs
    def hdfs_inotify_settings(self, zone=None):
        """Get HDFS inotify settings for an access zone."""
        try:
            hdfs_inotify_settings = self._sdk.ProtocolsApi(
                self._api_client,
            ).get_hdfs_inotify_settings(
                zone=zone or self.default_zone,
            ).settings
        except AttributeError:
            raise UnsupportedOperation('OneFS 8.1.1 or later is required for INotify support.')
        return {
            'enabled': hdfs_inotify_settings.enabled,
            'maximum_delay': hdfs_inotify_settings.maximum_delay,
            'retention': hdfs_inotify_settings.retention,
        }

    @accesses_onefs
    def hdfs_settings(self, zone=None):
        """Get HDFS settings for an access zone."""
        hdfs_settings = self._sdk.ProtocolsApi(self._api_client).get_hdfs_settings(
            zone=zone or self.default_zone,
        ).settings
        return {
            'ambari_namenode': hdfs_settings.ambari_namenode,
            'ambari_server': hdfs_settings.ambari_server,
            'authentication_mode': hdfs_settings.authentication_mode,
            'default_block_size': hdfs_settings.default_block_size,
            'default_checksum_type': hdfs_settings.default_checksum_type,
            'odp_version': hdfs_settings.odp_version,
            'root_directory': hdfs_settings.root_directory,
            'service': hdfs_settings.service,
            'webhdfs_enabled': hdfs_settings.webhdfs_enabled,
        }

    @property
    def host(self):
        """Get the URL to connect to OneFS at."""
        return self._configuration.host

    @host.setter
    def host(self, host):
        """Set the URL to connect to OneFS at."""
        self._configuration.host = host
        # self.host may now point to an unrelated address:
        self._address = None
        # self.host may now point to a different version of OneFS:
        self._refresh_sdk()

    def list_spns(self, provider):
        """Get a list of keytab entries for a Kerberos auth provider."""
        return [keytab_entry.spn for keytab_entry in self._keytab_entries(provider=provider)]

    @accesses_onefs
    def mkdir(self, path, mode, recursive=False, overwrite=False, zone=None):
        """Create a directory at a (zone-root-relative) path with the given (integer) mode."""
        real_path = self._zone_real_path(path, zone=zone or self.default_zone)
        if posixpath.sep not in real_path.strip(posixpath.sep):
            # The first component of real_path is actually a RAN namespace.
            # In this case, there is only one component: ifs.
            # The ifs namespace cannot be modified,
            # but calling create_directory on any namespace will fail.
            raise OneFSValueError('Calling mkdir on the ifs namespace will fail.')
        self._sdk.NamespaceApi(self._api_client).create_directory(
            directory_path=real_path.lstrip(posixpath.sep),
            x_isi_ifs_target_type='container',
            x_isi_ifs_access_control='{:o}'.format(mode),
            recursive=recursive,
            overwrite=overwrite,
        )

    @accesses_onefs
    def node_addresses(self, zone=None):
        """Get IP addresses in pools associated with a zone."""
        return {
            socket.inet_ntoa(struct.pack("!I", ip))
            for pool in self._sdk.NetworkApi(self._api_client).get_network_pools(
                access_zone=zone or self.default_zone,
            ).pools
            for range_ in pool.ranges
            for ip in range(
                struct.unpack("!I", socket.inet_aton(range_.low))[0],
                struct.unpack("!I", socket.inet_aton(range_.high))[0] + 1,
            )
        }

    @property
    def password(self):
        """Get the password to connect to OneFS with."""
        return self._configuration.password

    @password.setter
    def password(self, password):
        """Set the password to connect to OneFS with."""
        self._configuration.password = password

    @accesses_onefs
    def permissions(self, path, zone=None):
        """Get the owner, group, and (integer) mode of a (zone-root-relative) path."""
        real_path = self._zone_real_path(path, zone=zone or self.default_zone)
        acl = self._sdk.NamespaceApi(self._api_client).get_acl(
            namespace_path=real_path.lstrip(posixpath.sep),
            acl=True,
        )
        return {
            'group': acl.group.name,
            'mode': int(acl.mode, base=8),
            'owner': acl.owner.name,
        }

    @accesses_onefs
    def primary_group_of_user(self, user_name, zone=None):
        """Get the name of the primary group of a user."""
        auth_users = self._sdk.AuthApi(self._api_client).get_auth_user(
            user_name,
            zone=zone or self.default_zone,
        )
        assert len(auth_users.users) == 1, 'Do you have duplicate users (e.g. local and LDAP)?'
        return auth_users.users[0].gid.name

    def realms(self):
        """Get the Kerberos realms OneFS knows about."""
        for realm in self._realms():
            yield realm.realm

    def revision(self):
        """Get the revision number of the cluster."""
        revisions = set(self.revisions().values())
        if len(revisions) != 1:
            raise MixedModeError(', '.join(revisions))
        return revisions.pop()

    def revisions(self):
        """Get the revision numbers of each node in the cluster."""
        return {node.id: int(node.revision) for node in self._version().nodes}

    @accesses_onefs
    def rmdir(self, path, recursive=False, zone=None):
        """Delete the directory at a (zone-root-relative) path."""
        real_path = self._zone_real_path(path, zone=zone or self.default_zone)
        self._sdk.NamespaceApi(self._api_client).delete_directory(
            directory_path=real_path.lstrip(posixpath.sep),
            recursive=recursive,
        )

    def smartconnect_zone(self, smartconnect):
        """Get the access zone name associated with a SmartConnect name."""
        for pool in self._pools():
            if pool.sc_dns_zone.lower() == smartconnect.lower():
                return pool.access_zone
        return None

    @accesses_onefs
    def uid_of_user(self, user_name, zone=None):
        """Get the UID of a user."""
        auth_users = self._sdk.AuthApi(self._api_client).get_auth_user(
            user_name,
            zone=zone or self.default_zone,
        )
        assert len(auth_users.users) == 1, 'Do you have duplicate users (e.g. local and LDAP)?'
        return int(auth_users.users[0].uid.id.split(':')[1])

    @accesses_onefs
    def update_acl_settings(self, settings):
        """Set ACL settings."""
        acl_settings = self._sdk.SettingsAclsAclPolicySettings()
        for key, value in settings.items():
            try:
                getattr(acl_settings, key)
            except AttributeError as exc:
                raise_from(OneFSValueError('"{0}" is not a valid ACL setting.'.format(key)), exc)
            setattr(acl_settings, key, value)
        self._sdk.AuthApi(self._api_client).update_settings_acls(acl_settings)

    @accesses_onefs
    def update_hdfs_settings(self, settings, zone=None):
        """Set HDFS settings for an access zone."""
        hdfs_settings = self._sdk.HdfsSettingsSettings()
        for key, value in settings.items():
            try:
                getattr(hdfs_settings, key)
            except AttributeError as exc:
                raise_from(OneFSValueError('"{0}" is not a valid HDFS setting.'.format(key)), exc)
            setattr(hdfs_settings, key, value)
        self._sdk.ProtocolsApi(self._api_client).update_hdfs_settings(
            hdfs_settings,
            zone=zone or self.default_zone,
        )

    @accesses_onefs
    def update_zone_settings(self, settings, zone=None):
        """Set the settings for an access zone."""
        zone_settings = self._sdk.Zone()
        for key, value in settings.items():
            try:
                getattr(zone_settings, key)
            except AttributeError as exc:
                raise_from(OneFSValueError('"{0}" is not a valid zone setting.'.format(key)), exc)
            setattr(zone_settings, key, value)
        self._sdk.ZonesApi(self._api_client).update_zone(zone_settings, zone or self.default_zone)

    @property
    def username(self):
        """Get the user to connect to OneFS as."""
        return self._configuration.username

    @username.setter
    def username(self, username):
        """Set the user to connect to OneFS as."""
        self._configuration.username = username

    @accesses_onefs
    def user_groups(self, user_name, zone=None):
        """Get the groups a user is in."""
        auth_users = self._sdk.AuthApi(self._api_client).get_auth_user(
            auth_user_id='USER:{name}'.format(name=user_name),
            query_member_of=True,
            zone=zone or self.default_zone,
        )
        assert len(auth_users.users) == 1, 'Do you have duplicate users (e.g. local and LDAP)?'
        return [group.name for group in auth_users.users[0].member_of]

    @accesses_onefs
    def users(
            self,
            zone=None,
            key=lambda sdk_auth_user: sdk_auth_user.name,
            filter_=lambda _: True,
    ):
        """Get a list of users that exist in an access zone on OneFS."""
        for user in self._sdk.AuthApi(self._api_client).list_auth_users(
                zone=zone or self.default_zone,
        ).users:
            if filter_(user):
                yield key(user)

    @property
    def verify_ssl(self):
        """Determine whether the OneFS SSL certificate will be verified or not."""
        return self._configuration.verify_ssl

    @verify_ssl.setter
    def verify_ssl(self, verify_ssl):
        """Specify whether to verify the OneFS SSL certificate or not."""
        self._configuration.verify_ssl = bool(verify_ssl)

    @property
    def zone(self):
        """Get the default zone (used when a zone is not provided for a zone-specific operation)."""
        return self.default_zone

    def zone_settings(self, zone=None):
        """Get settings for an access zone."""
        zone_settings = self._zone(zone or self.default_zone)
        return {
            'alternate_system_provider': zone_settings.alternate_system_provider,
            'auth_providers': zone_settings.auth_providers,
            'cache_entry_expiry': zone_settings.cache_entry_expiry,
            'create_path': zone_settings.create_path,
            'groupnet': zone_settings.groupnet,
            'home_directory_umask': zone_settings.home_directory_umask,
            'id': zone_settings.id,
            'map_untrusted': zone_settings.map_untrusted,
            'name': zone_settings.name,
            'netbios_name': zone_settings.netbios_name,
            'path': zone_settings.path,
            'skeleton_directory': zone_settings.skeleton_directory,
            'system': zone_settings.system,
            'system_provider': zone_settings.system_provider,
            'user_mapping_rules': zone_settings.user_mapping_rules,
            'zone_id': zone_settings.zone_id,
        }

    def zones(self):
        """Get the list of access zone names available on the cluster."""
        for zone in self._zones():
            yield zone.name


class Client(BaseClient):

    """Do some basic checks after connecting to OneFS."""

    @classmethod
    def for_hdfs(cls, *args, **kwargs):
        """Connect to OneFS and do some basic HDFS-related checks."""
        onefs = cls(*args, **kwargs)
        onefs.check_license('HDFS')
        LOGGER.debug('HDFS is licensed.')
        return onefs

    def __init__(self, *args, **kwargs):
        LOGGER.debug('Connecting to the OneFS cluster...')
        super(Client, self).__init__(*args, **kwargs)
        LOGGER.debug('OneFS interactions will go to %s.', self.host)
        self.check_zone(self.zone)
        LOGGER.debug('The %s zone exists.', self.zone)



"""Verify the functionality of isilon_hadoop_tools.cli."""


from __future__ import absolute_import
from __future__ import unicode_literals

try:
    from unittest.mock import Mock  # Python 3
except ImportError:
    from mock import Mock  # Python 2

import pytest

from isilon_hadoop_tools import IsilonHadoopToolError, cli


def test_catches(exception):
    """Ensure cli.catches detects the desired exception."""
    assert cli.catches(exception)(Mock(side_effect=exception))() == 1


def test_not_catches(exception):
    """Ensure cli.catches does not catch undesirable exceptions."""
    with pytest.raises(exception):
        cli.catches(())(Mock(side_effect=exception))()


@pytest.mark.parametrize(
    'error, classinfo',
    [
        (cli.CLIError, IsilonHadoopToolError),
        (cli.HintedError, cli.CLIError),
    ],
)
def test_errors_cli(error, classinfo):
    """Ensure that exception types remain consistent."""
    assert issubclass(error, IsilonHadoopToolError)
    assert issubclass(error, cli.CLIError)
    assert issubclass(error, classinfo)



"""Isilon Hadoop Tools"""


from pkg_resources import get_distribution


__all__ = [
    # Constants
    '__version__',
    # Exceptions
    'IsilonHadoopToolError',
]
__version__ = get_distribution(__name__).version


class IsilonHadoopToolError(Exception):
    """All Exceptions emitted from this package inherit from this Exception."""

    def __str__(self):
        return super(IsilonHadoopToolError, self).__str__() or repr(self)

    def __repr__(self):
        return '{0}{cause}'.format(
            super(IsilonHadoopToolError, self).__repr__(),
            cause=(
                ' caused by {0!r}'.format(self.__cause__)  # pylint: disable=no-member
                if getattr(self, '__cause__', None) else
                ''
            ),
        )



#!/usr/bin/env python
# coding: utf-8

"""Packaging for Isilon Hadoop Tools"""

import setuptools

with open('README.md') as readme_file:
    README = readme_file.read()

setuptools.setup(
    name='isilon_hadoop_tools',
    use_scm_version=True,
    description='Tools for Using Hadoop with OneFS',
    long_description=README,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/isilon/isilon_hadoop_tools',
    maintainer='Isilon',
    maintainer_email='support@isilon.com',
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    include_package_data=True,
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    setup_requires=['setuptools_scm ~= 3.3'],
    install_requires=[
        'enum34 >= 1.1.6; python_version<"3.4"',
        'future >= 0.16.0',
        'isi-sdk-7-2 ~= 0.2.11',
        'isi-sdk-8-0 ~= 0.2.11',
        'isi-sdk-8-0-1 ~= 0.2.11',
        'isi-sdk-8-1-0 ~= 0.2.11',
        'isi-sdk-8-1-1 ~= 0.2.11',
        'isi-sdk-8-2-0 ~= 0.2.11',
        'isi-sdk-8-2-1 ~= 0.2.11',
        'isi-sdk-8-2-2 ~= 0.2.11',
        'requests >= 2.20.0',
        'setuptools >= 41.0.0',
    ],
    entry_points={
        'console_scripts': [
            'isilon_create_directories = isilon_hadoop_tools._scripts:isilon_create_directories',
            'isilon_create_users = isilon_hadoop_tools._scripts:isilon_create_users',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



"""Define and create directories with appropriate permissions on OneFS."""

from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import posixpath

import isilon_hadoop_tools.onefs
from isilon_hadoop_tools import IsilonHadoopToolError

__all__ = [
    # Exceptions
    'DirectoriesError',
    'HDFSRootDirectoryError',

    # Functions
    'cdh_directories',
    'cdp_directories',
    'hdp_directories',

    # Objects
    'Creator',
    'HDFSDirectory',
]

LOGGER = logging.getLogger(__name__)


class DirectoriesError(IsilonHadoopToolError):
    """All exceptions emitted from this module inherit from this Exception."""


class HDFSRootDirectoryError(DirectoriesError):
    """This exception occurs when the HDFS root directory is not set to a usable path."""


class Creator(object):

    """Create directories with appropriate ownership and permissions on OneFS."""

    def __init__(self, onefs, onefs_zone=None):
        self.onefs = onefs
        self.onefs_zone = onefs_zone

    def create_directories(self, directories, setup=None, mkdir=None, chmod=None, chown=None):
        """Create directories on HDFS on OneFS."""
        if self.onefs_zone.lower() == 'system':
            LOGGER.warning('Deploying in the System zone is not recommended.')
        sep = posixpath.sep
        zone_root = self.onefs.zone_settings(zone=self.onefs_zone)['path'].rstrip(sep)
        hdfs_root = self.onefs.hdfs_settings(zone=self.onefs_zone)['root_directory'].rstrip(sep)
        if hdfs_root == zone_root:
            LOGGER.warning('The HDFS root is the same as the zone root.')
        if hdfs_root == '/ifs':
            # The HDFS root requires non-default ownership/permissions,
            # and modifying /ifs can break NFS/SMB.
            raise HDFSRootDirectoryError(hdfs_root)
        assert hdfs_root.startswith(zone_root)
        zone_hdfs = hdfs_root[len(zone_root):]
        if setup:
            setup(zone_root, hdfs_root, zone_hdfs)
        for directory in directories:
            path = posixpath.join(zone_hdfs, directory.path.lstrip(posixpath.sep))
            LOGGER.info("mkdir '%s%s'", zone_root, path)
            try:
                (mkdir or self.onefs.mkdir)(path, directory.mode, zone=self.onefs_zone)
            except isilon_hadoop_tools.onefs.APIError as exc:
                if exc.dir_path_already_exists_error():
                    LOGGER.warning("%s%s already exists. ", zone_root, path)
                else:
                    raise
            LOGGER.info("chmod '%o' '%s%s'", directory.mode, zone_root, path)
            (chmod or self.onefs.chmod)(path, directory.mode, zone=self.onefs_zone)
            LOGGER.info("chown '%s:%s' '%s%s'", directory.owner, directory.group, zone_root, path)
            (chown or self.onefs.chown)(
                path,
                owner=directory.owner,
                group=directory.group,
                zone=self.onefs_zone,
            )

    def log_directories(self, directories):
        """Log the actions that would be taken by create_directories."""
        def _pass(*_, **__):
            pass
        self.create_directories(directories, setup=_pass, mkdir=_pass, chmod=_pass, chown=_pass)


class HDFSDirectory(object):  # pylint: disable=too-few-public-methods

    """A Directory on HDFS"""

    def __init__(self, path, owner, group, mode):
        self.path = path
        self.owner = owner
        self.group = group
        self.mode = mode

    def apply_identity_suffix(self, suffix):
        """Append a suffix to all identities associated with the directory."""
        self.owner += suffix
        self.group += suffix


def cdh_directories(identity_suffix=None):
    """Directories needed for Cloudera Distribution including Hadoop"""
    directories = [
        HDFSDirectory('/', 'hdfs', 'hadoop', 0o755),
        HDFSDirectory('/hbase', 'hbase', 'hbase', 0o755),
        HDFSDirectory('/solr', 'solr', 'solr', 0o775),
        HDFSDirectory('/tmp', 'hdfs', 'supergroup', 0o1777),
        HDFSDirectory('/tmp/hive', 'hive', 'supergroup', 0o777),
        HDFSDirectory('/tmp/logs', 'mapred', 'hadoop', 0o1777),
        HDFSDirectory('/user', 'hdfs', 'supergroup', 0o755),
        HDFSDirectory('/user/flume', 'flume', 'flume', 0o775),
        HDFSDirectory('/user/hdfs', 'hdfs', 'hdfs', 0o755),
        HDFSDirectory('/user/history', 'mapred', 'hadoop', 0o777),
        HDFSDirectory('/user/hive', 'hive', 'hive', 0o775),
        HDFSDirectory('/user/hive/warehouse', 'hive', 'hive', 0o1777),
        HDFSDirectory('/user/hue', 'hue', 'hue', 0o755),
        HDFSDirectory('/user/hue/.cloudera_manager_hive_metastore_canary', 'hue', 'hue', 0o777),
        HDFSDirectory('/user/impala', 'impala', 'impala', 0o775),
        HDFSDirectory('/user/oozie', 'oozie', 'oozie', 0o775),
        HDFSDirectory('/user/spark', 'spark', 'spark', 0o751),
        HDFSDirectory('/user/spark/applicationHistory', 'spark', 'spark', 0o1777),
        HDFSDirectory('/user/sqoop2', 'sqoop2', 'sqoop', 0o775),
        HDFSDirectory('/user/yarn', 'yarn', 'yarn', 0o755),
    ]
    if identity_suffix:
        for directory in directories:
            directory.apply_identity_suffix(identity_suffix)
    return directories


def cdp_directories(identity_suffix=None):
    """Directories needed for Cloudera Data Platform"""
    directories = [
        HDFSDirectory('/', 'hdfs', 'hadoop', 0o755),
        HDFSDirectory('/hbase', 'hbase', 'hbase', 0o755),
        HDFSDirectory('/ranger', 'hdfs', 'supergroup', 0o755),
        HDFSDirectory('/ranger/audit', 'hdfs', 'supergroup', 0o755),
        HDFSDirectory('/solr', 'solr', 'solr', 0o775),
        HDFSDirectory('/tmp', 'hdfs', 'supergroup', 0o1777),
        HDFSDirectory('/tmp/hive', 'hive', 'supergroup', 0o777),
        HDFSDirectory('/tmp/logs', 'yarn', 'hadoop', 0o1777),
        HDFSDirectory('/user', 'hdfs', 'supergroup', 0o755),
        HDFSDirectory('/user/flume', 'flume', 'flume', 0o775),
        HDFSDirectory('/user/hdfs', 'hdfs', 'hdfs', 0o755),
        HDFSDirectory('/user/history', 'mapred', 'hadoop', 0o777),
        HDFSDirectory('/user/history/done_intermediate', 'mapred', 'hadoop', 0o1777),
        HDFSDirectory('/user/hive', 'hive', 'hive', 0o775),
        HDFSDirectory('/user/hive/warehouse', 'hive', 'hive', 0o1777),
        HDFSDirectory('/user/hue', 'hue', 'hue', 0o755),
        HDFSDirectory('/user/hue/.cloudera_manager_hive_metastore_canary', 'hue', 'hue', 0o777),
        HDFSDirectory('/user/impala', 'impala', 'impala', 0o775),
        HDFSDirectory('/user/livy', 'livy', 'livy', 0o775),
        HDFSDirectory('/user/oozie', 'oozie', 'oozie', 0o775),
        HDFSDirectory('/user/spark', 'spark', 'spark', 0o751),
        HDFSDirectory('/user/spark/applicationHistory', 'spark', 'spark', 0o1777),
        HDFSDirectory('/user/spark/driverLogs', 'spark', 'spark', 0o1777),
        HDFSDirectory('/user/sqoop', 'sqoop', 'sqoop', 0o775),
        HDFSDirectory('/user/sqoop2', 'sqoop2', 'sqoop', 0o775),
        HDFSDirectory('/user/tez', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/user/yarn', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/user/yarn/mapreduce', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/user/yarn/mapreduce/mr-framework', 'yarn', 'hadoop', 0o775),
        HDFSDirectory('/user/yarn/services', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/user/yarn/services/service-framework', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/user/zeppelin', 'zeppelin', 'zeppelin', 0o775),
        HDFSDirectory('/warehouse', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/warehouse/tablespace', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/warehouse/tablespace/external', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/warehouse/tablespace/managed', 'hdfs', 'supergroup', 0o775),
        HDFSDirectory('/warehouse/tablespace/external/hive', 'hive', 'hive', 0o1775),
        HDFSDirectory('/warehouse/tablespace/managed/hive', 'hive', 'hive', 0o1775),
    ]
    if identity_suffix:
        for directory in directories:
            directory.apply_identity_suffix(identity_suffix)
    return directories


def hdp_directories(identity_suffix=None):
    """Directories needed for Hortonworks Data Platform"""
    directories = [
        HDFSDirectory('/', 'hdfs', 'hadoop', 0o755),
        HDFSDirectory('/app-logs', 'yarn', 'hadoop', 0o1777),
        HDFSDirectory('/app-logs/ambari-qa', 'ambari-qa', 'hadoop', 0o770),
        HDFSDirectory('/app-logs/ambari-qa/logs', 'ambari-qa', 'hadoop', 0o770),
        HDFSDirectory('/apps', 'hdfs', 'hadoop', 0o755),
        HDFSDirectory('/apps/accumulo', 'accumulo', 'hadoop', 0o750),
        HDFSDirectory('/apps/falcon', 'falcon', 'hdfs', 0o777),
        HDFSDirectory('/apps/hbase', 'hdfs', 'hadoop', 0o755),
        HDFSDirectory('/apps/hbase/data', 'hbase', 'hadoop', 0o775),
        HDFSDirectory('/apps/hbase/staging', 'hbase', 'hadoop', 0o711),
        HDFSDirectory('/apps/hive', 'hdfs', 'hdfs', 0o755),
        HDFSDirectory('/apps/hive/warehouse', 'hive', 'hdfs', 0o777),
        HDFSDirectory('/apps/tez', 'tez', 'hdfs', 0o755),
        HDFSDirectory('/apps/webhcat', 'hcat', 'hdfs', 0o755),
        HDFSDirectory('/ats', 'yarn', 'hdfs', 0o755),
        HDFSDirectory('/ats/done', 'yarn', 'hdfs', 0o775),
        HDFSDirectory('/atsv2', 'yarn-ats', 'hadoop', 0o755),
        HDFSDirectory('/mapred', 'mapred', 'hadoop', 0o755),
        HDFSDirectory('/mapred/system', 'mapred', 'hadoop', 0o755),
        HDFSDirectory('/system', 'yarn', 'hadoop', 0o755),
        HDFSDirectory('/system/yarn', 'yarn', 'hadoop', 0o755),
        HDFSDirectory('/system/yarn/node-labels', 'yarn', 'hadoop', 0o700),
        HDFSDirectory('/tmp', 'hdfs', 'hdfs', 0o1777),
        HDFSDirectory('/tmp/hive', 'ambari-qa', 'hdfs', 0o777),
        HDFSDirectory('/user', 'hdfs', 'hdfs', 0o755),
        HDFSDirectory('/user/ambari-qa', 'ambari-qa', 'hdfs', 0o770),
        HDFSDirectory('/user/hcat', 'hcat', 'hdfs', 0o755),
        HDFSDirectory('/user/hdfs', 'hdfs', 'hdfs', 0o755),
        HDFSDirectory('/user/hive', 'hive', 'hdfs', 0o700),
        HDFSDirectory('/user/hue', 'hue', 'hue', 0o755),
        HDFSDirectory('/user/oozie', 'oozie', 'hdfs', 0o775),
        HDFSDirectory('/user/yarn', 'yarn', 'hdfs', 0o755),
    ]
    if identity_suffix:
        for directory in directories:
            directory.apply_identity_suffix(identity_suffix)
    return directories



"""Verify the functionality of isilon_hadoop_tools.onefs."""


from __future__ import absolute_import
from __future__ import unicode_literals

import socket
try:
    # Python 3
    from unittest.mock import Mock
    from urllib.parse import urlparse
except ImportError:
    # Python 2
    from mock import Mock
    from urlparse import urlparse
import uuid

import isi_sdk_7_2
import isi_sdk_8_0
import isi_sdk_8_0_1
import isi_sdk_8_1_0
import isi_sdk_8_1_1
import isi_sdk_8_2_0
import isi_sdk_8_2_1
import isi_sdk_8_2_2
import pytest

from isilon_hadoop_tools import IsilonHadoopToolError, onefs


def test_init_connection_error(invalid_address, pytestconfig):
    """Creating a Client for an unusable host should raise a OneFSConnectionError."""
    with pytest.raises(onefs.OneFSConnectionError):
        onefs.Client(address=invalid_address, username=None, password=None)


@pytest.mark.xfail(
    raises=onefs.MalformedAPIError,
    reason='https://bugs.west.isilon.com/show_bug.cgi?id=248011',
)
def test_init_bad_creds(pytestconfig):
    """Creating a Client with invalid credentials should raise an appropriate exception."""
    with pytest.raises(onefs.APIError):
        onefs.Client(
            address=pytestconfig.getoption('--address', skip=True),
            username=str(uuid.uuid4()),
            password=str(uuid.uuid4()),
            verify_ssl=False,  # OneFS uses a self-signed certificate by default.
        )


def test_init(request):
    """Creating a Client should not raise an Exception."""
    assert isinstance(request.getfixturevalue('onefs_client'), onefs.Client)


def test_api_error_errors(api_error_errors_expectation):
    """Verify that APIError.errors raises appropriate exceptions."""
    api_error, expectation = api_error_errors_expectation
    with expectation:
        api_error.errors()


def test_api_error_str(api_error):
    """Verify that APIErrors can be stringified."""
    assert isinstance(str(api_error), str)


@pytest.mark.parametrize(
    'revision, expected_sdk',
    [
        (0, isi_sdk_8_2_2),
        (onefs.ONEFS_RELEASES['7.2.0.0'], isi_sdk_7_2),
        (onefs.ONEFS_RELEASES['8.0.0.0'], isi_sdk_8_0),
        (onefs.ONEFS_RELEASES['8.0.0.4'], isi_sdk_8_0),
        (onefs.ONEFS_RELEASES['8.0.1.0'], isi_sdk_8_0_1),
        (onefs.ONEFS_RELEASES['8.0.1.1'], isi_sdk_8_0_1),
        (onefs.ONEFS_RELEASES['8.1.0.0'], isi_sdk_8_1_0),
        (onefs.ONEFS_RELEASES['8.1.1.0'], isi_sdk_8_1_1),
        (onefs.ONEFS_RELEASES['8.1.2.0'], isi_sdk_8_1_1),
        (onefs.ONEFS_RELEASES['8.2.0.0'], isi_sdk_8_2_0),
        (onefs.ONEFS_RELEASES['8.2.1.0'], isi_sdk_8_2_1),
        (onefs.ONEFS_RELEASES['8.2.2.0'], isi_sdk_8_2_2),
        (onefs.ONEFS_RELEASES['8.2.3.0'], isi_sdk_8_2_2),
        (float('inf'), isi_sdk_8_2_2),
    ],
)
def test_sdk_for_revision(revision, expected_sdk):
    """Verify that an appropriate SDK is selected for a given revision."""
    assert onefs.sdk_for_revision(revision) is expected_sdk


def test_sdk_for_revision_unsupported():
    """Ensure that an UnsupportedVersion exception for unsupported revisions."""
    with pytest.raises(onefs.UnsupportedVersion):
        onefs.sdk_for_revision(revision=0, strict=True)


def test_accesses_onefs_connection_error(max_retry_exception_mock, onefs_client):
    """Verify that MaxRetryErrors are converted to OneFSConnectionErrors."""
    with pytest.raises(onefs.OneFSConnectionError):
        onefs.accesses_onefs(max_retry_exception_mock)(onefs_client)


def test_accesses_onefs_api_error(empty_api_exception_mock, onefs_client):
    """Verify that APIExceptions are converted to APIErrors."""
    with pytest.raises(onefs.APIError):
        onefs.accesses_onefs(empty_api_exception_mock)(onefs_client)


def test_accesses_onefs_try_again(retriable_api_exception_mock, onefs_client):
    """Verify that APIExceptions are retried appropriately."""
    mock, return_value = retriable_api_exception_mock
    assert onefs.accesses_onefs(mock)(onefs_client) == return_value


def test_accesses_onefs_other(exception, onefs_client):
    """Verify that arbitrary exceptions are not caught."""
    with pytest.raises(exception):
        onefs.accesses_onefs(Mock(side_effect=exception))(onefs_client)


def test_address(onefs_client, pytestconfig):
    """Verify that onefs.Client.address is exactly what was passed in."""
    assert onefs_client.address == pytestconfig.getoption('--address')


def test_username(onefs_client, pytestconfig):
    """Verify that onefs.Client.username is exactly what was passed in."""
    assert onefs_client.username == pytestconfig.getoption('--username')


def test_password(onefs_client, pytestconfig):
    """Verify that onefs.Client.password is exactly what was passed in."""
    assert onefs_client.password == pytestconfig.getoption('--password')


def test_host(onefs_client):
    """Verify that onefs.Client.host is a parsable url."""
    parsed = urlparse(onefs_client.host)
    assert parsed.scheme == 'https'
    assert socket.gethostbyname(parsed.hostname)
    assert parsed.port == 8080


def test_create_group(request):
    """Ensure that a group can be created successfully."""
    request.getfixturevalue('created_group')


def test_delete_group(onefs_client, deletable_group):
    """Verify that a group can be deleted successfully."""
    group_name, _ = deletable_group
    assert onefs_client.delete_group(name=group_name) is None


def test_gid_of_group(onefs_client, created_group):
    """Verify that the correct GID is fetched for an existing group."""
    group_name, gid = created_group
    assert onefs_client.gid_of_group(group_name=group_name) == gid


def test_groups(onefs_client, created_group):
    """Verify that a group that is known to exist appears in the list of existing groups."""
    group_name, _ = created_group
    assert group_name in onefs_client.groups()


def test_delete_user(onefs_client, deletable_user):
    """Verify that a user can be deleted successfully."""
    assert onefs_client.delete_user(name=deletable_user[0]) is None


def test_create_user(request):
    """Ensure that a user can be created successfully."""
    request.getfixturevalue('created_user')


def test_add_user_to_group(onefs_client, created_user, created_group):
    """Ensure that a user can be added to a group successfully."""
    assert onefs_client.add_user_to_group(
        user_name=created_user[0],
        group_name=created_group[0],
    ) is None


def test_create_hdfs_proxy_user(request):
    """Ensure that an HDFS proxy user can be created successfully."""
    request.getfixturevalue('created_proxy_user')


def test_delete_proxy_user(onefs_client, deletable_proxy_user):
    """Verify that a proxy user can be deleted successfully."""
    assert onefs_client.delete_hdfs_proxy_user(name=deletable_proxy_user[0]) is None


def test_uid_of_user(onefs_client, created_user):
    """Verify that the correct UID is fetched for an existing user."""
    user_name, _, uid = created_user
    assert onefs_client.uid_of_user(user_name=user_name) == uid


def test_primary_group_of_user(onefs_client, created_user):
    """Verify that the correct primary group is fetched for an existing user."""
    user_name, primary_group, _ = created_user
    assert onefs_client.primary_group_of_user(user_name=user_name) == primary_group


def test_create_realm(request):
    """Verify that a Kerberos realm can be created successfully."""
    request.getfixturevalue('created_realm')


def test_delete_realm(onefs_client, deletable_realm):
    """Verify that a realm can be deleted successfully."""
    onefs_client.delete_realm(name=deletable_realm)


def test_create_auth_provider(request):
    """Verify that a Kerberos auth provider can be created successfully."""
    request.getfixturevalue('created_auth_provider')


def test_delete_auth_provider(onefs_client, deletable_auth_provider):
    """Verify that a Kerberos auth provider can be deleted successfully."""
    onefs_client.delete_auth_provider(name=deletable_auth_provider)


def test_delete_spn(onefs_client, deletable_spn):
    """Verify that an SPN can be deleted successfully."""
    spn, provider = deletable_spn
    onefs_client.delete_spn(spn=spn, provider=provider)


def test_create_spn(request):
    """Verify that a Kerberos SPN can be created successfully."""
    request.getfixturevalue('created_spn')


def test_list_spns(onefs_client, created_spn):
    """Verify that a Kerberos SPN can be listed successfully."""
    spn, provider = created_spn
    assert (spn + '@' + provider) in onefs_client.list_spns(provider=provider)


def test_flush_auth_cache(onefs_client):
    """Verify that flushing the auth cache does not raise an exception."""
    assert onefs_client.flush_auth_cache() is None


def test_flush_auth_cache_unsupported(riptide_client):
    """
    Verify that trying flush the auth cache of a non-System zone
    before Halfpipe raises an UnsupportedOperation exception.
    """
    with pytest.raises(onefs.UnsupportedOperation):
        riptide_client.flush_auth_cache(zone='notSystem')


@pytest.mark.usefixtures('requests_delete_raises')
def test_flush_auth_cache_error(riptide_client):
    """
    Verify that flushing the auth cache raises an appropriate exception
    when things go wrong before Halfpipe.
    """
    with pytest.raises(onefs.NonSDKAPIError):
        riptide_client.flush_auth_cache()


def test_hdfs_inotify_settings(onefs_client):
    """Ensure hdfs_inotify_settings returns all available settings appropriately."""
    try:
        hdfs_inotify_settings = onefs_client.hdfs_inotify_settings()
    except onefs.UnsupportedOperation:
        assert onefs_client.revision() < onefs.ONEFS_RELEASES['8.1.1.0']
    else:
        assert isinstance(hdfs_inotify_settings, dict)
        assert all(
            setting in hdfs_inotify_settings
            for setting in ['enabled', 'maximum_delay', 'retention']
        )


@pytest.mark.parametrize(
    'setting_and_type',
    {
        'alternate_system_provider': str,
        'auth_providers': list,
        'cache_entry_expiry': int,
        'create_path': (bool, type(None)),
        'groupnet': str,
        'home_directory_umask': int,
        'id': str,
        'map_untrusted': str,
        'name': str,
        'netbios_name': str,
        'path': str,
        'skeleton_directory': str,
        'system': bool,
        'system_provider': str,
        'user_mapping_rules': list,
        'zone_id': int,
    }.items(),
)
def test_zone_settings(onefs_client, setting_and_type):
    """Ensure zone_settings returns all available settings appropriately."""
    setting, setting_type = setting_and_type
    assert isinstance(onefs_client.zone_settings()[setting], setting_type)


def test_zone_settings_bad_zone(onefs_client):
    """Ensure zone_settings fails appropriately when given a nonexistent zone."""
    with pytest.raises(onefs.MissingZoneError):
        onefs_client.zone_settings(zone=str(uuid.uuid4()))


def test_mkdir(request, onefs_client):
    """Ensure that a directory can be created successfully."""
    path, permissions = request.getfixturevalue('created_directory')

    def _check_postconditions():
        assert onefs_client.permissions(path) == permissions
    request.addfinalizer(_check_postconditions)


@pytest.mark.parametrize('recursive', [False, True])
def test_rmdir(onefs_client, deletable_directory, recursive, request):
    """Verify that a directory can be deleted successfully."""
    path, _ = deletable_directory
    assert onefs_client.rmdir(path=path, recursive=recursive) is None

    def _check_postconditions():
        with pytest.raises(onefs.APIError):
            onefs_client.permissions(path)
    request.addfinalizer(_check_postconditions)


def test_permissions(onefs_client, created_directory):
    """Check that permissions returns correct information."""
    path, permissions = created_directory
    assert onefs_client.permissions(path) == permissions


def test_chmod(onefs_client, created_directory, max_mode, request):
    """Check that chmod modifies the mode correctly."""
    path, permissions = created_directory
    new_mode = (permissions['mode'] + 1) % (max_mode + 1)
    assert onefs_client.chmod(path, new_mode) is None

    def _check_postconditions():
        assert onefs_client.permissions(path)['mode'] == new_mode
    request.addfinalizer(_check_postconditions)


@pytest.mark.parametrize('new_owner', [True, False])
@pytest.mark.parametrize('new_group', [True, False])
def test_chown(
        onefs_client,
        created_directory,
        created_user,
        created_group,
        new_owner,
        new_group,
        request,
):
    """Check that chown modifies ownership correctly."""
    path, permissions = created_directory
    user_name = created_user[0]
    group_name = created_group[0]
    assert onefs_client.chown(
        path,
        owner=user_name if new_owner else None,
        group=group_name if new_group else None,
    ) is None

    def _check_postconditions():
        owner = user_name if new_owner else permissions['owner']
        assert onefs_client.permissions(path)['owner'] == owner
        group = group_name if new_group else permissions['group']
        assert onefs_client.permissions(path)['group'] == group
    request.addfinalizer(_check_postconditions)


def test_feature_supported(onefs_client, supported_feature):
    """Ensure that feature_is_supported correctly identifies a supported feature."""
    try:
        assert onefs_client.feature_is_supported(supported_feature)
    except onefs.UnsupportedOperation:
        assert onefs_client.revision() < onefs.ONEFS_RELEASES['8.2.0.0']


def test_feature_unsupported(onefs_client, unsupported_feature):
    """Ensure that feature_is_supported correctly identifies an unsupported feature."""
    try:
        assert not onefs_client.feature_is_supported(unsupported_feature)
    except onefs.UnsupportedOperation:
        assert onefs_client.revision() < onefs.ONEFS_RELEASES['8.2.0.0']


@pytest.mark.parametrize(
    'error, classinfo',
    [
        (onefs.APIError, onefs.OneFSError),
        (onefs.ExpiredLicenseError, onefs.MissingLicenseError),
        (onefs.MalformedAPIError, onefs.OneFSError),
        (onefs.MissingLicenseError, onefs.OneFSError),
        (onefs.MissingZoneError, onefs.OneFSError),
        (onefs.MixedModeError, onefs.OneFSError),
        (onefs.OneFSCertificateError, onefs.OneFSConnectionError),
        (onefs.OneFSConnectionError, onefs.OneFSError),
        (onefs.OneFSError, IsilonHadoopToolError),
        (onefs.OneFSValueError, ValueError),
        (onefs.NonSDKAPIError, onefs.OneFSError),
        (onefs.UndecodableAPIError, onefs.MalformedAPIError),
        (onefs.UndeterminableVersion, onefs.OneFSError),
        (onefs.UnsupportedOperation, onefs.OneFSError),
        (onefs.UnsupportedVersion, onefs.OneFSError),
    ],
)
def test_errors_onefs(error, classinfo):
    """Ensure that exception types remain consistent."""
    assert issubclass(error, IsilonHadoopToolError)
    assert issubclass(error, onefs.OneFSError)
    assert issubclass(error, classinfo)
